from opcua import ua
from opcua.ua import uatypes
from enum import IntEnum

class ExampleEnum(IntEnum):
    EnumVal1 = 0
    EnumVal2 = 1
    EnumVal3 = 2

import opcua.ua
setattr(opcua.ua, 'ExampleEnum', ExampleEnum)

class ExampleStruct(uatypes.FrozenClass):

    ua_types = [
        ('IntVal1', 'Int16'),
        ('EnumVal', 'ExampleEnum'),
               ]

    def __init__(self):
        self.IntVal1 = 0
        self.EnumVal = ExampleEnum(0)
        self._freeze = True

    def __str__(self):
        return 'ExampleStruct(' + 'IntVal1:' + str(self.IntVal1) + ', ' + \
               'EnumVal:' + str(self.EnumVal) + ')'

    __repr__ = __str__

def add_server_custom_enum_struct(server):
    # import some nodes from xml
    server.import_xml("enum_struct_test_nodes.xml")
    ns = server.get_namespace_index('http://yourorganisation.org/struct_enum_example/')
    uatypes.register_extension_object('ExampleStruct', ua.NodeId(5001, ns), ExampleStruct)
    val = ua.ExampleStruct()
    val.IntVal1 = 242
    val.EnumVal = ua.ExampleEnum.EnumVal2
    myvar = server.get_node(ua.NodeId(6009, ns))
    myvar.set_value(val)



# encoding: utf-8
from concurrent.futures import Future, TimeoutError
import time
from datetime import datetime
from datetime import timedelta
import math
from contextlib import contextmanager

from opcua import ua
from opcua import Node
from opcua import uamethod
from opcua import instantiate
from opcua import copy_node
from opcua.common import ua_utils
from opcua.common.methods import call_method_full

def add_server_methods(srv):
    @uamethod
    def func(parent, value):
        return value * 2

    o = srv.get_objects_node()
    v = o.add_method(ua.NodeId("ServerMethod", 2), ua.QualifiedName('ServerMethod', 2), func, [ua.VariantType.Int64], [ua.VariantType.Int64])

    @uamethod
    def func2(parent, methodname, value):
        if methodname == "panic":
            return ua.StatusCode(ua.StatusCodes.BadOutOfMemory)
        if methodname != "sin":
            res = ua.CallMethodResult()
            res.StatusCode = ua.StatusCode(ua.StatusCodes.BadInvalidArgument)
            res.InputArgumentResults = [ua.StatusCode(ua.StatusCodes.BadNotSupported), ua.StatusCode()]
            return res
        return math.sin(value)

    o = srv.get_objects_node()
    v = o.add_method(ua.NodeId("ServerMethodArray", 2), ua.QualifiedName('ServerMethodArray', 2), func2, [ua.VariantType.String, ua.VariantType.Int64], [ua.VariantType.Int64])

    @uamethod
    def func3(parent, mylist):
        return [i * 2 for i in mylist]

    o = srv.get_objects_node()
    v = o.add_method(ua.NodeId("ServerMethodArray2", 2), ua.QualifiedName('ServerMethodArray2', 2), func3, [ua.VariantType.Int64], [ua.VariantType.Int64])

    @uamethod
    def func4(parent):
        return None

    base_otype= srv.get_node(ua.ObjectIds.BaseObjectType)
    custom_otype = base_otype.add_object_type(2, 'ObjectWithMethodsType')
    custom_otype.add_method(2, 'ServerMethodDefault', func4)
    custom_otype.add_method(2, 'ServerMethodMandatory', func4).set_modelling_rule(True)
    custom_otype.add_method(2, 'ServerMethodOptional', func4).set_modelling_rule(False)
    custom_otype.add_method(2, 'ServerMethodNone', func4).set_modelling_rule(None)
    o.add_object(2, 'ObjectWithMethods', custom_otype)

    @uamethod
    def func5(parent):
        return 1,2,3
    o = srv.get_objects_node()
    v = o.add_method(ua.NodeId("ServerMethodTuple", 2), ua.QualifiedName('ServerMethodTuple', 2), func5, [], [ua.VariantType.Int64, ua.VariantType.Int64, ua.VariantType.Int64])



class CommonTests(object):

    '''
    Tests that will be run twice. Once on server side and once on
    client side since we have been carefull to have the exact
    same api on server and client side
    '''
    # jyst to avoid editor warnings
    opc = None
    assertEqual = lambda x, y: True
    assertIn = lambda x, y: True

    @contextmanager
    def assertNotRaises(self, exc_type):
        try:
            yield None
        except exc_type:
            raise self.failureException('{} raised'.format(exc_type.__name__))

    def test_find_servers(self):
        servers = self.opc.find_servers()
        # FIXME : finish

    def test_add_node_bad_args(self):
        obj = self.opc.get_objects_node()

        with self.assertRaises(TypeError):
            fold = obj.add_folder(1.2, "kk")

        with self.assertRaises(TypeError):
            fold = obj.add_folder(ua.UaError, "khjh")

        with self.assertRaises(ua.UaError):
            fold = obj.add_folder("kjk", 1.2)

        with self.assertRaises(TypeError):
            fold = obj.add_folder("i=0;s='oooo'", 1.2)

        with self.assertRaises(ua.UaError):
            fold = obj.add_folder("i=0;s='oooo'", "tt:oioi")

    def test_delete_nodes(self):
        obj = self.opc.get_objects_node()
        fold = obj.add_folder(2, "FolderToDelete")
        var = fold.add_variable(2, "VarToDelete", 9.1)
        childs = fold.get_children()
        self.assertIn(var, childs)
        self.opc.delete_nodes([var])
        with self.assertRaises(ua.UaStatusCodeError):
            var.set_value(7.8)
        with self.assertRaises(ua.UaStatusCodeError):
            obj.get_child(["2:FolderToDelete", "2:VarToDelete"])
        childs = fold.get_children()
        self.assertNotIn(var, childs)


    def test_delete_nodes_with_inverse_references(self):
        obj = self.opc.get_objects_node()
        fold = obj.add_folder(2, "FolderToDelete")
        var = fold.add_variable(2, "VarToDelete", 9.1)
        var2 = fold.add_variable(2, "VarWithReference", 9.2)
        childs = fold.get_children()
        self.assertIn(var, childs)
        self.assertIn(var2, childs)
        # add two references to var, this includes adding the inverse references to var2
        var.add_reference(var2.nodeid, reftype=ua.ObjectIds.HasDescription, forward=True, bidirectional=True)
        var.add_reference(var2.nodeid, reftype=ua.ObjectIds.HasEffect, forward=True, bidirectional=True)
        self.opc.delete_nodes([var])
        childs = fold.get_children()
        assert var not in childs
        has_desc_refs = var2.get_referenced_nodes(refs=ua.ObjectIds.HasDescription, direction=ua.BrowseDirection.Inverse)
        assert len(has_desc_refs) == 0
        has_effect_refs = var2.get_referenced_nodes(refs=ua.ObjectIds.HasEffect, direction=ua.BrowseDirection.Inverse)
        assert len(has_effect_refs) == 0


    def test_delete_nodes_recursive(self):
        obj = self.opc.get_objects_node()
        fold = obj.add_folder(2, "FolderToDeleteR")
        var = fold.add_variable(2, "VarToDeleteR", 9.1)
        self.opc.delete_nodes([fold, var])
        with self.assertRaises(ua.UaStatusCodeError):
            var.set_value(7.8)
        with self.assertRaises(ua.UaStatusCodeError):
            obj.get_child(["2:FolderToDelete", "2:VarToDelete"])

    def test_delete_nodes_recursive2(self):
        obj = self.opc.get_objects_node()
        fold = obj.add_folder(2, "FolderToDeleteRoot")
        nfold = fold
        mynodes = []
        for i in range(7):
            nfold = fold.add_folder(2, "FolderToDeleteRoot")
            var = fold.add_variable(2, "VarToDeleteR", 9.1)
            var = fold.add_property(2, "ProToDeleteR", 9.1)
            prop = fold.add_property(2, "ProToDeleteR", 9.1)
            o = fold.add_object(3, "ObjToDeleteR")
            o_var = o.add_variable(3, "VarToDeleteRR", 9.2)
            o_prop = o.add_property(3, "PropToDeleteRR", 9.2)
            mynodes.append(nfold)
            mynodes.append(var)
            mynodes.append(prop)
            mynodes.append(o)
            mynodes.append(o_var)
            mynodes.append(o_prop)
        self.opc.delete_nodes([fold], recursive=True)
        for node in mynodes:
            with self.assertRaises(ua.UaStatusCodeError):
                node.get_browse_name()

    def test_delete_references(self):
        newtype = self.opc.get_node(ua.ObjectIds.HierarchicalReferences).add_reference_type(0, "HasSuperSecretVariable")

        obj = self.opc.get_objects_node()
        fold = obj.add_folder(2, "FolderToRef")
        var = fold.add_variable(2, "VarToRef", 42)

        fold.add_reference(var, newtype)

        self.assertEqual(var.get_referenced_nodes(newtype), [fold])
        self.assertEqual(fold.get_referenced_nodes(newtype), [var])

        fold.delete_reference(var, newtype)

        self.assertEqual(var.get_referenced_nodes(newtype), [])
        self.assertEqual(fold.get_referenced_nodes(newtype), [])

        fold.add_reference(var, newtype, bidirectional=False)

        self.assertEqual(var.get_referenced_nodes(newtype), [])
        self.assertEqual(fold.get_referenced_nodes(newtype), [var])

        fold.delete_reference(var, newtype)

        self.assertEqual(var.get_referenced_nodes(newtype), [])
        self.assertEqual(fold.get_referenced_nodes(newtype), [])


        var.add_reference(fold, newtype, forward=False, bidirectional=False)

        self.assertEqual(var.get_referenced_nodes(newtype), [fold])
        self.assertEqual(fold.get_referenced_nodes(newtype), [])

        with self.assertRaises(ua.UaStatusCodeError):
            fold.delete_reference(var, newtype)

        self.assertEqual(var.get_referenced_nodes(newtype), [fold])
        self.assertEqual(fold.get_referenced_nodes(newtype), [])

        with self.assertRaises(ua.UaStatusCodeError):
            var.delete_reference(fold, newtype)

        self.assertEqual(var.get_referenced_nodes(newtype), [fold])
        self.assertEqual(fold.get_referenced_nodes(newtype), [])

        var.delete_reference(fold, newtype, forward=False)

        self.assertEqual(var.get_referenced_nodes(newtype), [])
        self.assertEqual(fold.get_referenced_nodes(newtype), [])

        # clean-up
        self.opc.delete_nodes([fold, newtype], recursive=True)

    def test_server_node(self):
        node = self.opc.get_server_node()
        self.assertEqual(ua.QualifiedName('Server', 0), node.get_browse_name())

    def test_root(self):
        root = self.opc.get_root_node()
        self.assertEqual(ua.QualifiedName('Root', 0), root.get_browse_name())
        self.assertEqual(ua.LocalizedText('Root'), root.get_display_name())
        nid = ua.NodeId(84, 0)
        self.assertEqual(nid, root.nodeid)

    def test_objects(self):
        objects = self.opc.get_objects_node()
        self.assertEqual(ua.QualifiedName('Objects', 0), objects.get_browse_name())
        nid = ua.NodeId(85, 0)
        self.assertEqual(nid, objects.nodeid)

    def test_browse(self):
        objects = self.opc.get_objects_node()
        obj = objects.add_object(4, "browsetest")
        folder = obj.add_folder(4, "folder")
        prop = obj.add_property(4, "property", 1)
        prop2 = obj.add_property(4, "property2", 2)
        var = obj.add_variable(4, "variable", 3)
        obj2 = obj.add_object(4, "obj")
        alle = obj.get_children()
        self.assertTrue(prop in alle)
        self.assertTrue(prop2 in alle)
        self.assertTrue(var in alle)
        self.assertTrue(folder in alle)
        self.assertFalse(obj in alle)
        props = obj.get_children(refs=ua.ObjectIds.HasProperty)
        self.assertTrue(prop in props)
        self.assertTrue(prop2 in props)
        self.assertFalse(var in props)
        self.assertFalse(folder in props)
        self.assertFalse(obj2 in props)
        all_vars = obj.get_children(nodeclassmask=ua.NodeClass.Variable)
        self.assertTrue(prop in all_vars)
        self.assertTrue(var in all_vars)
        self.assertFalse(folder in props)
        self.assertFalse(obj2 in props)
        all_objs = obj.get_children(nodeclassmask=ua.NodeClass.Object)
        self.assertTrue(folder in all_objs)
        self.assertTrue(obj2 in all_objs)
        self.assertFalse(var in all_objs)

    def test_browse_references(self):
        objects = self.opc.get_objects_node()
        folder = objects.add_folder(4, "folder")

        childs = objects.get_referenced_nodes(refs=ua.ObjectIds.Organizes, direction=ua.BrowseDirection.Forward, includesubtypes=False)
        self.assertTrue(folder in childs)

        childs = objects.get_referenced_nodes(refs=ua.ObjectIds.Organizes, direction=ua.BrowseDirection.Both, includesubtypes=False)
        self.assertTrue(folder in childs)

        childs = objects.get_referenced_nodes(refs=ua.ObjectIds.Organizes, direction=ua.BrowseDirection.Inverse, includesubtypes=False)
        self.assertFalse(folder in childs)

        parents = folder.get_referenced_nodes(refs=ua.ObjectIds.Organizes, direction=ua.BrowseDirection.Inverse, includesubtypes=False)
        self.assertTrue(objects in parents)

        parents = folder.get_referenced_nodes(refs=ua.ObjectIds.HierarchicalReferences, direction=ua.BrowseDirection.Inverse, includesubtypes=False)
        self.assertTrue(objects in parents)

        parent = folder.get_parent()
        self.assertEqual(parent, objects)

    def test_browsename_with_spaces(self):
        o = self.opc.get_objects_node()
        v = o.add_variable(3, 'BNVariable with spaces and %&+?/', 1.3)
        v2 = o.get_child("3:BNVariable with spaces and %&+?/")
        self.assertEqual(v, v2)

    def test_non_existing_path(self):
        root = self.opc.get_root_node()
        with self.assertRaises(ua.UaStatusCodeError):
            server_time_node = root.get_child(['0:Objects', '0:Server', '0:nonexistingnode'])

    def test_bad_attribute(self):
        root = self.opc.get_root_node()
        with self.assertRaises(ua.UaStatusCodeError):
            root.set_value(99)

    def test_get_node_by_nodeid(self):
        root = self.opc.get_root_node()
        server_time_node = root.get_child(['0:Objects', '0:Server', '0:ServerStatus', '0:CurrentTime'])
        correct = self.opc.get_node(ua.NodeId(ua.ObjectIds.Server_ServerStatus_CurrentTime))
        self.assertEqual(server_time_node, correct)

    def test_datetime_read(self):
        time_node = self.opc.get_node(ua.NodeId(ua.ObjectIds.Server_ServerStatus_CurrentTime))
        dt = time_node.get_value()
        utcnow = datetime.utcnow()
        delta = utcnow - dt
        self.assertTrue(delta < timedelta(seconds=1))

    def test_datetime_write(self):
        time_node = self.opc.get_node(ua.NodeId(ua.ObjectIds.Server_ServerStatus_CurrentTime))
        now = datetime.utcnow()
        objects = self.opc.get_objects_node()
        v1 = objects.add_variable(4, "test_datetime", now)
        tid = v1.get_value()
        self.assertEqual(now, tid)

    def test_variant_array_dim(self):
        objects = self.opc.get_objects_node()
        l = [[[1.0, 1.0, 1.0, 1.0], [2.0, 2.0, 2.0, 2.0], [3.0, 3.0, 3.0, 3.0]],[[5.0, 5.0, 5.0, 5.0], [7.0, 8.0, 9.0, 01.0], [1.0, 1.0, 1.0, 1.0]]]
        v = objects.add_variable(3, 'variableWithDims', l)

        v.set_array_dimensions([0, 0, 0])
        dim = v.get_array_dimensions()
        self.assertEqual(dim, [0, 0, 0])

        v.set_value_rank(0)
        rank = v.get_value_rank()
        self.assertEqual(rank, 0)

        v2 = v.get_value()
        self.assertEqual(v2, l)
        dv = v.get_data_value()
        self.assertEqual(dv.Value.Dimensions, [2,3,4])

        l = [[[], [], []], [[], [], []]]
        variant = ua.Variant(l, ua.VariantType.UInt32)
        v = objects.add_variable(3, 'variableWithDimsEmpty', variant)
        v2 = v.get_value()
        self.assertEqual(v2, l)
        dv = v.get_data_value()
        self.assertEqual(dv.Value.Dimensions, [2,3,0])

    def test_add_numeric_variable(self):
        objects = self.opc.get_objects_node()
        v = objects.add_variable('ns=3;i=888;', '3:numericnodefromstring', 99)
        nid = ua.NodeId(888, 3)
        qn = ua.QualifiedName('numericnodefromstring', 3)
        self.assertEqual(nid, v.nodeid)
        self.assertEqual(qn, v.get_browse_name())

    def test_add_string_variable(self):
        objects = self.opc.get_objects_node()
        v = objects.add_variable('ns=3;s=stringid;', '3:stringnodefromstring', [68])
        nid = ua.NodeId('stringid', 3)
        qn = ua.QualifiedName('stringnodefromstring', 3)
        self.assertEqual(nid, v.nodeid)
        self.assertEqual(qn, v.get_browse_name())

    def test_utf8(self):
        objects = self.opc.get_objects_node()
        utf_string = "æøå@%&"
        bn = ua.QualifiedName(utf_string, 3)
        nid = ua.NodeId("æølå", 3)
        val = "æøå"
        v = objects.add_variable(nid, bn, val)
        self.assertEqual(nid, v.nodeid)
        val2 = v.get_value()
        self.assertEqual(val, val2)
        bn2 = v.get_browse_name()
        self.assertEqual(bn, bn2)

    def test_null_variable(self):
        objects = self.opc.get_objects_node()
        var = objects.add_variable(3, 'nullstring', "a string")
        var.set_value(None)
        val = var.get_value()
        self.assertEqual(val, None)
        var.set_value("")
        val = var.get_value()
        self.assertNotEqual(val, None)
        self.assertEqual(val, "")

    def test_variable_data_type(self):
        objects = self.opc.get_objects_node()
        var = objects.add_variable(3, 'stringfordatatype', "a string")
        val = var.get_data_type_as_variant_type()
        self.assertEqual(val, ua.VariantType.String)
        var = objects.add_variable(3, 'stringarrayfordatatype', ["a", "b"])
        val = var.get_data_type_as_variant_type()
        self.assertEqual(val, ua.VariantType.String)

    def test_add_string_array_variable(self):
        objects = self.opc.get_objects_node()
        v = objects.add_variable('ns=3;s=stringarrayid;', '9:stringarray', ['l', 'b'])
        nid = ua.NodeId('stringarrayid', 3)
        qn = ua.QualifiedName('stringarray', 9)
        self.assertEqual(nid, v.nodeid)
        self.assertEqual(qn, v.get_browse_name())
        val = v.get_value()
        self.assertEqual(['l', 'b'], val)

    def test_add_numeric_node(self):
        objects = self.opc.get_objects_node()
        nid = ua.NodeId(9999, 3)
        qn = ua.QualifiedName('AddNodeVar1', 3)
        v1 = objects.add_variable(nid, qn, 0)
        self.assertEqual(nid, v1.nodeid)
        self.assertEqual(qn, v1.get_browse_name())

    def test_add_string_node(self):
        objects = self.opc.get_objects_node()
        qn = ua.QualifiedName('AddNodeVar2', 3)
        nid = ua.NodeId('AddNodeVar2Id', 3)
        v2 = objects.add_variable(nid, qn, 0)
        self.assertEqual(nid, v2.nodeid)
        self.assertEqual(qn, v2.get_browse_name())

    def test_add_find_node_(self):
        objects = self.opc.get_objects_node()
        o = objects.add_object('ns=2;i=101;', '2:AddFindObject')
        o2 = objects.get_child('2:AddFindObject')
        self.assertEqual(o, o2)

    def test_node_path(self):
        objects = self.opc.get_objects_node()
        o = objects.add_object('ns=2;i=105;', '2:NodePathObject')
        root = self.opc.get_root_node()
        o2 = root.get_child(['0:Objects', '2:NodePathObject'])
        self.assertEqual(o, o2)

    def test_add_read_node(self):
        objects = self.opc.get_objects_node()
        o = objects.add_object('ns=2;i=102;', '2:AddReadObject')
        nid = ua.NodeId(102, 2)
        self.assertEqual(o.nodeid, nid)
        qn = ua.QualifiedName('AddReadObject', 2)
        self.assertEqual(o.get_browse_name(), qn)

    def test_simple_value(self):
        o = self.opc.get_objects_node()
        v = o.add_variable(3, 'VariableTestValue', 4.32)
        val = v.get_value()
        self.assertEqual(4.32, val)

    def test_add_exception(self):
        objects = self.opc.get_objects_node()
        o = objects.add_object('ns=2;i=103;', '2:AddReadObject')
        with self.assertRaises(ua.UaStatusCodeError):
            o2 = objects.add_object('ns=2;i=103;', '2:AddReadObject')

    def test_negative_value(self):
        o = self.opc.get_objects_node()
        v = o.add_variable(3, 'VariableNegativeValue', 4)
        v.set_value(-4.54)
        val = v.get_value()
        self.assertEqual(-4.54, val)

    def test_read_server_state(self):
        statenode = self.opc.get_node(ua.NodeId(ua.ObjectIds.Server_ServerStatus_State))
        state = statenode.get_value()
        self.assertEqual(state, 0)

    def test_bad_node(self):
        bad = self.opc.get_node(ua.NodeId(999, 999))
        with self.assertRaises(ua.UaStatusCodeError):
            bad.get_browse_name()
        with self.assertRaises(ua.UaStatusCodeError):
            bad.set_value(89)
        with self.assertRaises(ua.UaStatusCodeError):
            bad.add_object(0, "0:myobj")
        with self.assertRaises(ua.UaStatusCodeError):
            bad.get_child("0:myobj")

    def test_value(self):
        o = self.opc.get_objects_node()
        var = ua.Variant(1.98, ua.VariantType.Double)
        v = o.add_variable(3, 'VariableValue', var)
        val = v.get_value()
        self.assertEqual(1.98, val)

        dvar = ua.DataValue(var)
        dv = v.get_data_value()
        self.assertEqual(ua.DataValue, type(dv))
        self.assertEqual(dvar.Value, dv.Value)
        self.assertEqual(dvar.Value, var)

    def test_set_value(self):
        o = self.opc.get_objects_node()
        var = ua.Variant(1.98, ua.VariantType.Double)
        dvar = ua.DataValue(var)
        v = o.add_variable(3, 'VariableValue', var)
        v.set_value(var.Value)
        v1 = v.get_value()
        self.assertEqual(v1, var.Value)
        v.set_value(var)
        v2 = v.get_value()
        self.assertEqual(v2, var.Value)
        v.set_data_value(dvar)
        v3 = v.get_data_value()
        self.assertEqual(v3.Value, dvar.Value)

    def test_array_value(self):
        o = self.opc.get_objects_node()
        v = o.add_variable(3, 'VariableArrayValue', [1, 2, 3])
        val = v.get_value()
        self.assertEqual([1, 2, 3], val)

    def test_bool_variable(self):
        o = self.opc.get_objects_node()
        v = o.add_variable(3, 'BoolVariable', True)
        dt = v.get_data_type_as_variant_type()
        self.assertEqual(dt, ua.VariantType.Boolean)
        val = v.get_value()
        self.assertEqual(True, val)
        v.set_value(False)
        val = v.get_value()
        self.assertEqual(False, val)

    def test_array_size_one_value(self):
        o = self.opc.get_objects_node()
        v = o.add_variable(3, 'VariableArrayValue', [1, 2, 3])
        v.set_value([1])
        val = v.get_value()
        self.assertEqual([1], val)

    def test_use_namespace(self):
        idx_orig = 1
        uri = self.opc.get_namespace_array()[idx_orig]
        idx = self.opc.get_namespace_index(uri)
        self.assertEqual(idx_orig, idx)
        root = self.opc.get_root_node()
        myvar = root.add_variable(idx, 'var_in_custom_namespace', [5])
        myid = myvar.nodeid
        self.assertEqual(idx, myid.NamespaceIndex)

    def test_method(self):
        o = self.opc.get_objects_node()
        m = o.get_child("2:ServerMethod")
        result = o.call_method("2:ServerMethod", 2.1)
        self.assertEqual(result, 4.2)
        with self.assertRaises(ua.UaStatusCodeError):
            # FIXME: we should raise a more precise exception
            result = o.call_method("2:ServerMethod", 2.1, 89, 9)
        with self.assertRaises(ua.UaStatusCodeError):
            result = o.call_method(ua.NodeId(999), 2.1)  # non existing method

    def test_method_array(self):
        o = self.opc.get_objects_node()
        m = o.get_child("2:ServerMethodArray")
        result = o.call_method(m, "sin", ua.Variant(math.pi))
        self.assertTrue(result < 0.01)
        with self.assertRaises(ua.UaStatusCodeError) as cm:
            result = o.call_method(m, "cos", ua.Variant(math.pi))
        self.assertEqual(cm.exception.code, ua.StatusCodes.BadInvalidArgument)
        with self.assertRaises(ua.UaStatusCodeError) as cm:
            result = o.call_method(m, "panic", ua.Variant(math.pi))
        self.assertEqual(cm.exception.code, ua.StatusCodes.BadOutOfMemory)

    def test_method_array2(self):
        o = self.opc.get_objects_node()
        m = o.get_child("2:ServerMethodArray2")
        result = o.call_method(m, [1.1, 3.4, 9])
        self.assertEqual(result, [2.2, 6.8, 18])
        result = call_method_full(o, m, [1.1, 3.4, 9])
        self.assertEqual(result.OutputArguments, [[2.2, 6.8, 18]])

    def test_method_tuple(self):
        o = self.opc.get_objects_node()
        m = o.get_child("2:ServerMethodTuple")
        result = o.call_method(m)
        self.assertEqual(result, [1, 2, 3])
        result = call_method_full(o, m)
        self.assertEqual(result.OutputArguments, [1, 2, 3])

    def test_method_none(self):
        # this test calls the function linked to the type's method..
        o = self.opc.get_node(ua.ObjectIds.BaseObjectType).get_child("2:ObjectWithMethodsType")
        m = o.get_child("2:ServerMethodDefault")
        result = o.call_method(m)
        self.assertEqual(result, None)
        result = call_method_full(o, m)
        self.assertEqual(result.OutputArguments, [])

    def test_add_nodes(self):
        objects = self.opc.get_objects_node()
        f = objects.add_folder(3, 'MyFolder')
        child = objects.get_child("3:MyFolder")
        self.assertEqual(child, f)
        o = f.add_object(3, 'MyObject')
        child = f.get_child("3:MyObject")
        self.assertEqual(child, o)
        v = f.add_variable(3, 'MyVariable', 6)
        child = f.get_child("3:MyVariable")
        self.assertEqual(child, v)
        p = f.add_property(3, 'MyProperty', 10)
        child = f.get_child("3:MyProperty")
        self.assertEqual(child, p)
        childs = f.get_children()
        self.assertTrue(o in childs)
        self.assertTrue(v in childs)
        self.assertTrue(p in childs)

    def test_modelling_rules(self):
        obj = self.opc.nodes.base_object_type.add_object_type(2, 'MyFooObjectType')
        v = obj.add_variable(2, "myvar", 1.1)
        v.set_modelling_rule(True)
        p = obj.add_property(2, "myvar", 1.1)
        p.set_modelling_rule(False)

        refs = obj.get_referenced_nodes(refs=ua.ObjectIds.HasModellingRule)
        self.assertEqual(len(refs), 0)

        refs = v.get_referenced_nodes(refs=ua.ObjectIds.HasModellingRule)
        self.assertEqual(refs[0], self.opc.get_node(ua.ObjectIds.ModellingRule_Mandatory))

        refs = p.get_referenced_nodes(refs=ua.ObjectIds.HasModellingRule)
        self.assertEqual(refs[0], self.opc.get_node(ua.ObjectIds.ModellingRule_Optional))

        p.set_modelling_rule(None)
        refs = p.get_referenced_nodes(refs=ua.ObjectIds.HasModellingRule)
        self.assertEqual(len(refs), 0)

    def test_incl_subtypes(self):
        base_type = self.opc.get_root_node().get_child(["0:Types", "0:ObjectTypes", "0:BaseObjectType"])
        descs = base_type.get_children_descriptions(includesubtypes=True)
        self.assertTrue(len(descs) > 10)
        descs = base_type.get_children_descriptions(includesubtypes=False)
        self.assertEqual(len(descs), 0)

    def test_add_node_with_type(self):
        objects = self.opc.get_objects_node()
        f = objects.add_folder(3, 'MyFolder_TypeTest')

        o = f.add_object(3, 'MyObject1', ua.ObjectIds.BaseObjectType)
        self.assertEqual(o.get_type_definition().Identifier, ua.ObjectIds.BaseObjectType)

        o = f.add_object(3, 'MyObject2', ua.NodeId(ua.ObjectIds.BaseObjectType, 0))
        self.assertEqual(o.get_type_definition().Identifier, ua.ObjectIds.BaseObjectType)

        base_otype= self.opc.get_node(ua.ObjectIds.BaseObjectType)
        custom_otype = base_otype.add_object_type(2, 'MyFooObjectType')

        o = f.add_object(3, 'MyObject3', custom_otype.nodeid)
        self.assertEqual(o.get_type_definition().Identifier, custom_otype.nodeid.Identifier)

        references = o.get_references(refs=ua.ObjectIds.HasTypeDefinition, direction=ua.BrowseDirection.Forward)
        self.assertEqual(len(references), 1)
        self.assertEqual(references[0].NodeId, custom_otype.nodeid)

    def test_references_for_added_nodes(self):
        objects = self.opc.get_objects_node()
        o = objects.add_object(3, 'MyObject')
        nodes = objects.get_referenced_nodes(refs=ua.ObjectIds.Organizes, direction=ua.BrowseDirection.Forward, includesubtypes=False)
        self.assertTrue(o in nodes)
        nodes = o.get_referenced_nodes(refs=ua.ObjectIds.Organizes, direction=ua.BrowseDirection.Inverse, includesubtypes=False)
        self.assertTrue(objects in nodes)
        self.assertEqual(o.get_parent(), objects)
        self.assertEqual(o.get_type_definition().Identifier, ua.ObjectIds.BaseObjectType)
        self.assertEqual(o.get_references(ua.ObjectIds.HasModellingRule), [])

        o2 = o.add_object(3, 'MySecondObject')
        nodes = o.get_referenced_nodes(refs=ua.ObjectIds.HasComponent, direction=ua.BrowseDirection.Forward, includesubtypes=False)
        self.assertTrue(o2 in nodes)
        nodes = o2.get_referenced_nodes(refs=ua.ObjectIds.HasComponent, direction=ua.BrowseDirection.Inverse, includesubtypes=False)
        self.assertTrue(o in nodes)
        self.assertEqual(o2.get_parent(), o)
        self.assertEqual(o2.get_type_definition().Identifier, ua.ObjectIds.BaseObjectType)
        self.assertEqual(o2.get_references(ua.ObjectIds.HasModellingRule), [])

        v = o.add_variable(3, 'MyVariable', 6)
        nodes = o.get_referenced_nodes(refs=ua.ObjectIds.HasComponent, direction=ua.BrowseDirection.Forward, includesubtypes=False)
        self.assertTrue(v in nodes)
        nodes = v.get_referenced_nodes(refs=ua.ObjectIds.HasComponent, direction=ua.BrowseDirection.Inverse, includesubtypes=False)
        self.assertTrue(o in nodes)
        self.assertEqual(v.get_parent(), o)
        self.assertEqual(v.get_type_definition().Identifier, ua.ObjectIds.BaseDataVariableType)
        self.assertEqual(v.get_references(ua.ObjectIds.HasModellingRule), [])

        p = o.add_property(3, 'MyProperty', 2)
        nodes = o.get_referenced_nodes(refs=ua.ObjectIds.HasProperty, direction=ua.BrowseDirection.Forward, includesubtypes=False)
        self.assertTrue(p in nodes)
        nodes = p.get_referenced_nodes(refs=ua.ObjectIds.HasProperty, direction=ua.BrowseDirection.Inverse, includesubtypes=False)
        self.assertTrue(o in nodes)
        self.assertEqual(p.get_parent(), o)
        self.assertEqual(p.get_type_definition().Identifier, ua.ObjectIds.PropertyType)
        self.assertEqual(p.get_references(ua.ObjectIds.HasModellingRule), [])

        m = objects.get_child("2:ServerMethod")
        self.assertEqual(m.get_references(ua.ObjectIds.HasModellingRule), [])

    def test_path_string(self):
        o = self.opc.nodes.objects.add_folder(1, "titif").add_object(3, "opath")
        path = o.get_path(as_string=True)
        self.assertEqual(["0:Root", "0:Objects", "1:titif", "3:opath"], path)
        path = o.get_path(2, as_string=True)
        self.assertEqual(["1:titif", "3:opath"], path)

    def test_path(self):
        of = self.opc.nodes.objects.add_folder(1, "titif")
        op = of.add_object(3, "opath")
        path = op.get_path()
        self.assertEqual([self.opc.nodes.root, self.opc.nodes.objects, of, op], path)
        path = op.get_path(2)
        self.assertEqual([of, op], path)
        target = self.opc.get_node("i=13387")
        path = target.get_path()
        self.assertEqual([self.opc.nodes.root, self.opc.nodes.types, self.opc.nodes.object_types, self.opc.nodes.base_object_type, self.opc.nodes.folder_type, self.opc.get_node(ua.ObjectIds.FileDirectoryType), target], path)

    def test_get_endpoints(self):
        endpoints = self.opc.get_endpoints()
        self.assertTrue(len(endpoints) > 0)
        self.assertTrue(endpoints[0].EndpointUrl.startswith("opc.tcp://"))

    def test_copy_node(self):
        dev_t = self.opc.nodes.base_data_type.add_object_type(0, "MyDevice")
        v_t = dev_t.add_variable(0, "sensor", 1.0)
        p_t = dev_t.add_property(0, "sensor_id", "0340")
        ctrl_t = dev_t.add_object(0, "controller")
        prop_t = ctrl_t.add_property(0, "state", "Running")
        # Create device sutype
        devd_t = dev_t.add_object_type(0, "MyDeviceDervived")
        v_t = devd_t.add_variable(0, "childparam", 1.0)
        p_t = devd_t.add_property(0, "sensorx_id", "0340")


        nodes = copy_node(self.opc.nodes.objects, dev_t)
        mydevice = nodes[0]

        self.assertEqual(mydevice.get_node_class(), ua.NodeClass.ObjectType)
        self.assertEqual(len(mydevice.get_children()), 4)
        obj = mydevice.get_child(["0:controller"])
        prop = mydevice.get_child(["0:controller", "0:state"])
        self.assertEqual(prop.get_type_definition().Identifier, ua.ObjectIds.PropertyType)
        self.assertEqual(prop.get_value(), "Running")
        self.assertNotEqual(prop.nodeid, prop_t.nodeid)

    def test_instantiate_1(self):
        # Create device type
        dev_t = self.opc.nodes.base_object_type.add_object_type(0, "MyDevice")
        v_t = dev_t.add_variable(0, "sensor", 1.0)
        v_t.set_modelling_rule(True)
        p_t = dev_t.add_property(0, "sensor_id", "0340")
        p_t.set_modelling_rule(True)
        ctrl_t = dev_t.add_object(0, "controller")
        ctrl_t.set_modelling_rule(True)
        v_opt_t = dev_t.add_variable(0, "vendor", 1.0)
        v_opt_t.set_modelling_rule(False)
        v_none_t = dev_t.add_variable(0, "model", 1.0)
        v_none_t.set_modelling_rule(None)
        prop_t = ctrl_t.add_property(0, "state", "Running")
        prop_t.set_modelling_rule(True)

        # Create device sutype
        devd_t = dev_t.add_object_type(0, "MyDeviceDervived")
        v_t = devd_t.add_variable(0, "childparam", 1.0)
        v_t.set_modelling_rule(True)
        p_t = devd_t.add_property(0, "sensorx_id", "0340")
        p_t.set_modelling_rule(True)

        # instanciate device
        nodes = instantiate(self.opc.nodes.objects, dev_t, bname="2:Device0001")
        mydevice = nodes[0]

        self.assertEqual(mydevice.get_node_class(), ua.NodeClass.Object)
        self.assertEqual(mydevice.get_type_definition(), dev_t.nodeid)
        obj = mydevice.get_child(["0:controller"])
        prop = mydevice.get_child(["0:controller", "0:state"])
        with self.assertRaises(ua.UaError):
            mydevice.get_child(["0:controller", "0:vendor"])
        with self.assertRaises(ua.UaError):
            mydevice.get_child(["0:controller", "0:model"])

        self.assertEqual(prop.get_type_definition().Identifier, ua.ObjectIds.PropertyType)
        self.assertEqual(prop.get_value(), "Running")
        self.assertNotEqual(prop.nodeid, prop_t.nodeid)

        # instanciate device subtype
        nodes = instantiate(self.opc.nodes.objects, devd_t, bname="2:Device0002")
        mydevicederived = nodes[0]
        prop1 = mydevicederived.get_child(["0:sensorx_id"])
        var1 = mydevicederived.get_child(["0:childparam"])
        var_parent = mydevicederived.get_child(["0:sensor"])
        prop_parent = mydevicederived.get_child(["0:sensor_id"])

    def test_instantiate_string_nodeid(self):
        # Create device type
        dev_t = self.opc.nodes.base_object_type.add_object_type(0, "MyDevice2")
        v_t = dev_t.add_variable(0, "sensor", 1.0)
        v_t.set_modelling_rule(True)
        p_t = dev_t.add_property(0, "sensor_id", "0340")
        p_t.set_modelling_rule(True)
        ctrl_t = dev_t.add_object(0, "controller")
        ctrl_t.set_modelling_rule(True)
        prop_t = ctrl_t.add_property(0, "state", "Running")
        prop_t.set_modelling_rule(True)

        # instanciate device
        nodes = instantiate(self.opc.nodes.objects, dev_t, nodeid=ua.NodeId("InstDevice", 2, ua.NodeIdType.String), bname="2:InstDevice")
        mydevice = nodes[0]

        self.assertEqual(mydevice.get_node_class(), ua.NodeClass.Object)
        self.assertEqual(mydevice.get_type_definition(), dev_t.nodeid)
        obj = mydevice.get_child(["0:controller"])
        obj_nodeid_ident = obj.nodeid.Identifier
        prop = mydevice.get_child(["0:controller", "0:state"])
        self.assertEqual(obj_nodeid_ident, "InstDevice.controller")
        self.assertEqual(prop.get_type_definition().Identifier, ua.ObjectIds.PropertyType)
        self.assertEqual(prop.get_value(), "Running")
        self.assertNotEqual(prop.nodeid, prop_t.nodeid)

    def test_variable_with_datatype(self):
        v1 = self.opc.nodes.objects.add_variable(3, 'VariableEnumType1', ua.ApplicationType.ClientAndServer, datatype=ua.NodeId(ua.ObjectIds.ApplicationType))
        tp1 = v1.get_data_type()
        self.assertEqual(ua.NodeId(ua.ObjectIds.ApplicationType), tp1)

        v2 = self.opc.nodes.objects.add_variable(3, 'VariableEnumType2', ua.ApplicationType.ClientAndServer, datatype=ua.NodeId(ua.ObjectIds.ApplicationType) )
        tp2 = v2.get_data_type()
        self.assertEqual( ua.NodeId(ua.ObjectIds.ApplicationType), tp2)

    def test_enum(self):
        # create enum type
        enums = self.opc.get_root_node().get_child(["0:Types", "0:DataTypes", "0:BaseDataType", "0:Enumeration"])
        myenum_type = enums.add_data_type(0, "MyEnum")
        es = myenum_type.add_variable(0, "EnumStrings", [ua.LocalizedText("String0"), ua.LocalizedText("String1"), ua.LocalizedText("String2")], ua.VariantType.LocalizedText)
        #es.set_value_rank(1)
        # instantiate
        o = self.opc.get_objects_node()
        myvar = o.add_variable(2, "MyEnumVar", ua.LocalizedText("String1"), datatype=myenum_type.nodeid)
        #myvar.set_writable(True)
        # tests
        self.assertEqual(myvar.get_data_type(), myenum_type.nodeid)
        myvar.set_value(ua.LocalizedText("String2"))

    def test_supertypes(self):
        nint32 = self.opc.get_node(ua.ObjectIds.Int32)
        node = ua_utils.get_node_supertype(nint32)
        self.assertEqual(node, self.opc.get_node(ua.ObjectIds.Integer))

        nodes = ua_utils.get_node_supertypes(nint32)
        self.assertEqual(nodes[1], self.opc.get_node(ua.ObjectIds.Number))
        self.assertEqual(nodes[0], self.opc.get_node(ua.ObjectIds.Integer))

        # test custom
        dtype = nint32.add_data_type(0, "MyCustomDataType")
        node = ua_utils.get_node_supertype(dtype)
        self.assertEqual(node, nint32)

        dtype2 = dtype.add_data_type(0, "MyCustomDataType2")
        node = ua_utils.get_node_supertype(dtype2)
        self.assertEqual(node, dtype)

    def test_base_data_type(self):
        nint32 = self.opc.get_node(ua.ObjectIds.Int32)
        dtype = nint32.add_data_type(0, "MyCustomDataType")
        dtype2 = dtype.add_data_type(0, "MyCustomDataType2")
        self.assertEqual(ua_utils.get_base_data_type(dtype), nint32)
        self.assertEqual(ua_utils.get_base_data_type(dtype2), nint32)

        ext = self.opc.nodes.objects.add_variable(0, "MyExtensionObject", ua.Argument())
        d = ext.get_data_type()
        d = self.opc.get_node(d)
        self.assertEqual(ua_utils.get_base_data_type(d), self.opc.get_node(ua.ObjectIds.Structure))
        self.assertEqual(ua_utils.data_type_to_variant_type(d), ua.VariantType.ExtensionObject)

    def test_data_type_to_variant_type(self):
        test_data = {
            ua.ObjectIds.Boolean: ua.VariantType.Boolean,
            ua.ObjectIds.Byte: ua.VariantType.Byte,
            ua.ObjectIds.String: ua.VariantType.String,
            ua.ObjectIds.Int32: ua.VariantType.Int32,
            ua.ObjectIds.UInt32: ua.VariantType.UInt32,
            ua.ObjectIds.NodeId: ua.VariantType.NodeId,
            ua.ObjectIds.LocalizedText: ua.VariantType.LocalizedText,
            ua.ObjectIds.Structure: ua.VariantType.ExtensionObject,
            ua.ObjectIds.EnumValueType: ua.VariantType.ExtensionObject,
            ua.ObjectIds.Enumeration: ua.VariantType.Int32,  # enumeration
            ua.ObjectIds.AttributeWriteMask: ua.VariantType.UInt32,
            ua.ObjectIds.AxisScaleEnumeration: ua.VariantType.Int32  # enumeration
            }
        for dt, vdt in test_data.items():
            self.assertEqual(ua_utils.data_type_to_variant_type(self.opc.get_node(ua.NodeId(dt))), vdt)

    def test_guid_node_id(self):
        """
        Test that a Node can be instantiated with a GUID string and that the NodeId ca be converted to binary.
        """
        node = Node(None, "ns=4;g=35d5f86f-2777-4550-9d48-b098f5ee285c")
        binary_node_id = ua.ua_binary.nodeid_to_binary(node.nodeid)
        assert type(binary_node_id) is bytes



from concurrent.futures import Future, TimeoutError
import time
from datetime import datetime, timedelta
from copy import copy
import unittest

import opcua
from opcua import Client
from opcua import Server
from opcua import ua
from opcua.server.internal_server import InternalServer, InternalSession

try:
    from unittest import mock
except ImportError:
    import mock


class SubHandler():

    """
    Dummy subscription client
    """

    def datachange_notification(self, node, val, data):
        pass

    def event_notification(self, event):
        pass


class MySubHandler():

    """
    More advanced subscription client using Future, so we can wait for events in tests
    """

    def __init__(self):
        self.future = Future()

    def reset(self):
        self.future = Future()

    def datachange_notification(self, node, val, data):
        self.future.set_result((node, val, data))

    def event_notification(self, event):
        self.future.set_result(event)


class MySubHandler2():
    def __init__(self):
        self.results = []

    def datachange_notification(self, node, val, data):
        self.results.append((node, val))

    def event_notification(self, event):
        self.results.append(event)


class MySubHandlerCounter():
    def __init__(self):
        self.datachange_count = 0
        self.event_count = 0

    def datachange_notification(self, node, val, data):
        self.datachange_count += 1

    def event_notification(self, event):
        self.event_count += 1


class SubscriptionTests(object):

    def test_subscription_failure(self):
        myhandler = MySubHandler()
        o = self.opc.get_objects_node()
        sub = self.opc.create_subscription(100, myhandler)
        with self.assertRaises(ua.UaStatusCodeError):
            handle1 = sub.subscribe_data_change(o) # we can only subscribe to variables so this should fail
        sub.delete()

    def test_subscription_overload(self):
        nb = 10
        myhandler = MySubHandler()
        o = self.opc.get_objects_node()
        sub = self.opc.create_subscription(1, myhandler)
        variables = []
        subs = []
        for i in range(nb):
            v = o.add_variable(3, 'SubscriptionVariableOverload' + str(i), 99)
            variables.append(v)
        for i in range(nb):
            sub.subscribe_data_change(variables)
        for i in range(nb):
            for j in range(nb):
                variables[i].set_value(j)
            s = self.opc.create_subscription(1, myhandler)
            s.subscribe_data_change(variables)
            subs.append(s)
            sub.subscribe_data_change(variables[i])
        for i in range(nb):
            for j in range(nb):
                variables[i].set_value(j)
        sub.delete()
        for s in subs:
            s.delete()

    def test_subscription_count(self):
        myhandler = MySubHandlerCounter()
        sub = self.opc.create_subscription(1, myhandler)
        o = self.opc.get_objects_node()
        var = o.add_variable(3, 'SubVarCounter', 0.1)
        sub.subscribe_data_change(var)
        nb = 12
        for i in range(nb):
            val = var.get_value()
            var.set_value(val +1)
        time.sleep(0.2)  # let last event arrive
        self.assertEqual(myhandler.datachange_count, nb + 1)
        sub.delete()

    def test_subscription_count_list(self):
        myhandler = MySubHandlerCounter()
        sub = self.opc.create_subscription(1, myhandler)
        o = self.opc.get_objects_node()
        var = o.add_variable(3, 'SubVarCounter', [0.1, 0.2])
        sub.subscribe_data_change(var)
        nb = 12
        for i in range(nb):
            val = var.get_value()
            val = copy(val)  # we do not want to modify object in our db, we need a copy in order to generate event
            val.append(i)
            var.set_value(copy(val))
        time.sleep(0.2)  # let last event arrive
        self.assertEqual(myhandler.datachange_count, nb + 1)
        sub.delete()

    def test_subscription_count_no_change(self):
        myhandler = MySubHandlerCounter()
        sub = self.opc.create_subscription(1, myhandler)
        o = self.opc.get_objects_node()
        var = o.add_variable(3, 'SubVarCounter', [0.1, 0.2])
        sub.subscribe_data_change(var)
        nb = 12
        for i in range(nb):
            val = var.get_value()
            var.set_value(val)
        time.sleep(0.2)  # let last event arrive
        self.assertEqual(myhandler.datachange_count, 1)
        sub.delete()

    def test_subscription_count_empty(self):
        myhandler = MySubHandlerCounter()
        sub = self.opc.create_subscription(1, myhandler)
        o = self.opc.get_objects_node()
        var = o.add_variable(3, 'SubVarCounter', [0.1, 0.2, 0.3])
        sub.subscribe_data_change(var)
        while True:
            val = var.get_value()
            val = copy(val)  # we do not want to modify object in our db, we need a copy in order to generate event
            val.pop()
            var.set_value(val, ua.VariantType.Double)
            if not val:
                break
        time.sleep(0.2)  # let last event arrive
        self.assertEqual(myhandler.datachange_count, 4)
        sub.delete()

    def test_subscription_overload_simple(self):
        nb = 10
        myhandler = MySubHandler()
        o = self.opc.get_objects_node()
        sub = self.opc.create_subscription(1, myhandler)
        variables = [o.add_variable(3, 'SubVarOverload' + str(i), i) for i in range(nb)]
        for i in range(nb):
            sub.subscribe_data_change(variables)
        sub.delete()

    def test_subscription_data_change(self):
        """
        test subscriptions. This is far too complicated for
        a unittest but, setting up subscriptions requires a lot
        of code, so when we first set it up, it is best
        to test as many things as possible
        """
        myhandler = MySubHandler()

        o = self.opc.get_objects_node()

        # subscribe to a variable
        startv1 = [1, 2, 3]
        v1 = o.add_variable(3, 'SubscriptionVariableV1', startv1)
        sub = self.opc.create_subscription(100, myhandler)
        handle1 = sub.subscribe_data_change(v1)

        # Now check we get the start value
        node, val, data = myhandler.future.result()
        self.assertEqual(val, startv1)
        self.assertEqual(node, v1)

        myhandler.reset()  # reset future object

        # modify v1 and check we get value
        v1.set_value([5])
        node, val, data = myhandler.future.result()

        self.assertEqual(node, v1)
        self.assertEqual(val, [5])

        with self.assertRaises(ua.UaStatusCodeError):
            sub.unsubscribe(999)  # non existing handle
        sub.unsubscribe(handle1)
        with self.assertRaises(ua.UaStatusCodeError):
            sub.unsubscribe(handle1)  # second try should fail
        sub.delete()
        with self.assertRaises(ua.UaStatusCodeError):
            sub.unsubscribe(handle1)  # sub does not exist anymore

    def test_subscription_data_change_bool(self):
        """
        test subscriptions. This is far too complicated for
        a unittest but, setting up subscriptions requires a lot
        of code, so when we first set it up, it is best
        to test as many things as possible
        """
        myhandler = MySubHandler()

        o = self.opc.get_objects_node()

        # subscribe to a variable
        startv1 = True
        v1 = o.add_variable(3, 'SubscriptionVariableBool', startv1)
        sub = self.opc.create_subscription(100, myhandler)
        handle1 = sub.subscribe_data_change(v1)

        # Now check we get the start value
        node, val, data = myhandler.future.result()
        self.assertEqual(val, startv1)
        self.assertEqual(node, v1)

        myhandler.reset()  # reset future object

        # modify v1 and check we get value
        v1.set_value(False)
        node, val, data = myhandler.future.result()
        self.assertEqual(node, v1)
        self.assertEqual(val, False)

        sub.delete() # should delete our monitoreditem too

    def test_subscription_data_change_many(self):
        """
        test subscriptions. This is far too complicated for
        a unittest but, setting up subscriptions requires a lot
        of code, so when we first set it up, it is best
        to test as many things as possible
        """
        myhandler = MySubHandler2()
        o = self.opc.get_objects_node()

        startv1 = True
        v1 = o.add_variable(3, 'SubscriptionVariableMany1', startv1)
        startv2 = [1.22, 1.65]
        v2 = o.add_variable(3, 'SubscriptionVariableMany2', startv2)

        sub = self.opc.create_subscription(100, myhandler)
        handle1, handle2 = sub.subscribe_data_change([v1, v2])

        # Now check we get the start values
        nodes = [v1, v2]

        count = 0
        while not len(myhandler.results) > 1:
            count += 1
            time.sleep(0.1)
            if count > 100:
                self.fail("Did not get result from subscription")
        for node, val in myhandler.results:
            self.assertIn(node, nodes)
            nodes.remove(node)
            if node == v1:
                self.assertEqual(startv1, val)
            elif node == v2:
                self.assertEqual(startv2, val)
            else:
                self.fail("Error node {0} is neither {1} nor {2}".format(node, v1, v2))

        sub.delete()

    def test_subscribe_server_time(self):
        myhandler = MySubHandler()

        server_time_node = self.opc.get_node(ua.NodeId(ua.ObjectIds.Server_ServerStatus_CurrentTime))

        sub = self.opc.create_subscription(200, myhandler)
        handle = sub.subscribe_data_change(server_time_node)

        node, val, data = myhandler.future.result()
        self.assertEqual(node, server_time_node)
        delta = datetime.utcnow() - val
        self.assertTrue(delta < timedelta(seconds=2))

        sub.unsubscribe(handle)
        sub.delete()



    def test_create_delete_subscription(self):
        o = self.opc.get_objects_node()
        v = o.add_variable(3, 'SubscriptionVariable', [1, 2, 3])
        sub = self.opc.create_subscription(100, MySubHandler())
        handle = sub.subscribe_data_change(v)
        time.sleep(0.1)
        sub.unsubscribe(handle)
        sub.delete()

    def test_subscribe_events(self):
        sub = self.opc.create_subscription(100, MySubHandler())
        handle = sub.subscribe_events()
        time.sleep(0.1)
        sub.unsubscribe(handle)
        sub.delete()

    def test_subscribe_events_to_wrong_node(self):
        sub = self.opc.create_subscription(100, MySubHandler())
        with self.assertRaises(ua.UaStatusCodeError):
            handle = sub.subscribe_events(self.opc.get_node("i=85"))
        o = self.opc.get_objects_node()
        v = o.add_variable(3, 'VariableNoEventNofierAttribute', 4)
        with self.assertRaises(ua.UaStatusCodeError):
            handle = sub.subscribe_events(v)
        sub.delete()

    def test_get_event_from_type_node_BaseEvent(self):
        etype = self.opc.get_node(ua.ObjectIds.BaseEventType)
        properties = opcua.common.events.get_event_properties_from_type_node(etype)
        for child in etype.get_properties():
            self.assertTrue(child in properties)

    def test_get_event_from_type_node_CustomEvent(self):
        etype = self.srv.create_custom_event_type(2, 'MyEvent', ua.ObjectIds.AuditEventType, [('PropertyNum', ua.VariantType.Float), ('PropertyString', ua.VariantType.String)])

        properties = opcua.common.events.get_event_properties_from_type_node(etype)

        for child in self.opc.get_node(ua.ObjectIds.BaseEventType).get_properties():
            self.assertTrue(child in properties)
        for child in self.opc.get_node(ua.ObjectIds.AuditEventType).get_properties():
            self.assertTrue(child in properties)
        for child in self.opc.get_node(etype.nodeid).get_properties():
                self.assertTrue(child in properties)

        self.assertTrue(etype.get_child("2:PropertyNum") in properties)
        self.assertTrue(etype.get_child("2:PropertyString") in properties)

    def test_events_default(self):
        evgen = self.srv.get_event_generator()

        myhandler = MySubHandler()
        sub = self.opc.create_subscription(100, myhandler)
        handle = sub.subscribe_events()

        tid = datetime.utcnow()
        msg = "this is my msg "
        evgen.trigger(tid, msg)

        ev = myhandler.future.result()
        self.assertIsNot(ev, None)  # we did not receive event
        self.assertEqual(ev.EventType, ua.NodeId(ua.ObjectIds.BaseEventType))
        self.assertEqual(ev.Severity, 1)
        self.assertEqual(ev.SourceName, self.opc.get_server_node().get_browse_name().Name)
        self.assertEqual(ev.SourceNode, self.opc.get_server_node().nodeid)
        self.assertEqual(ev.Message.Text, msg)
        self.assertEqual(ev.Time, tid)

        # time.sleep(0.1)
        sub.unsubscribe(handle)
        sub.delete()

    def test_events_MyObject(self):
        objects = self.srv.get_objects_node()
        o = objects.add_object(3, 'MyObject')
        evgen = self.srv.get_event_generator()
        evgen.event.SourceNode=o.nodeid
        evgen.event.SourceName=o.get_browse_name().Name

        myhandler = MySubHandler()
        sub = self.opc.create_subscription(100, myhandler)
        handle = sub.subscribe_events()

        tid = datetime.utcnow()
        msg = "this is my msg "
        evgen.trigger(tid, msg)

        ev = myhandler.future.result(10)
        self.assertIsNot(ev, None)  # we did not receive event
        self.assertEqual(ev.EventType, ua.NodeId(ua.ObjectIds.BaseEventType))
        self.assertEqual(ev.Severity, 1)
        self.assertEqual(ev.SourceName, 'MyObject')
        self.assertEqual(ev.SourceNode, o.nodeid)
        self.assertEqual(ev.Message.Text, msg)
        self.assertEqual(ev.Time, tid)

        # time.sleep(0.1)
        sub.unsubscribe(handle)
        sub.delete()

    def test_events_wrong_source(self):
        objects = self.srv.get_objects_node()
        o = objects.add_object(3, 'MyObject')
        evgen = self.srv.get_event_generator(emitting_node=o)

        myhandler = MySubHandler()
        sub = self.opc.create_subscription(100, myhandler)
        handle = sub.subscribe_events()

        tid = datetime.utcnow()
        msg = "this is my msg "
        evgen.trigger(tid, msg)

        with self.assertRaises(TimeoutError):  # we should not receive event
            ev = myhandler.future.result(2)

        # time.sleep(0.1)
        sub.unsubscribe(handle)
        sub.delete()

    def test_events_CustomEvent(self):
        etype = self.srv.create_custom_event_type(2, 'MyEvent', ua.ObjectIds.BaseEventType, [('PropertyNum', ua.VariantType.Float), ('PropertyString', ua.VariantType.String)])
        evgen = self.srv.get_event_generator(etype)

        myhandler = MySubHandler()
        sub = self.opc.create_subscription(100, myhandler)
        handle = sub.subscribe_events(sourcenode=ua.ObjectIds.Server, evtypes=etype)

        propertynum = 2
        propertystring = "This is my test"
        evgen.event.PropertyNum = propertynum
        evgen.event.PropertyString = propertystring
        serverity = 500
        evgen.event.Severity = serverity
        tid = datetime.utcnow()
        msg = "this is my msg "
        evgen.trigger(tid, msg)

        ev = myhandler.future.result(10)
        self.assertIsNot(ev, None)  # we did not receive event
        self.assertEqual(ev.EventType, etype.nodeid)
        self.assertEqual(ev.Severity, serverity)
        self.assertEqual(ev.SourceName, self.opc.get_server_node().get_browse_name().Name)
        self.assertEqual(ev.SourceNode, self.opc.get_server_node().nodeid)
        self.assertEqual(ev.Message.Text, msg)
        self.assertEqual(ev.Time, tid)
        self.assertEqual(ev.PropertyNum, propertynum)
        self.assertEqual(ev.PropertyString, propertystring)

        # time.sleep(0.1)
        sub.unsubscribe(handle)
        sub.delete()

    def test_events_CustomEvent_MyObject(self):
        objects = self.srv.get_objects_node()
        o = objects.add_object(3, 'MyObject')
        etype = self.srv.create_custom_event_type(2, 'MyEvent', ua.ObjectIds.BaseEventType, [('PropertyNum', ua.VariantType.Float), ('PropertyString', ua.VariantType.String)])
        evgen = self.srv.get_event_generator(etype, emitting_node=o)
        evgen.event.SourceNode = o.nodeid

        myhandler = MySubHandler()
        sub = self.opc.create_subscription(100, myhandler)
        handle = sub.subscribe_events(sourcenode=o, evtypes=etype)

        propertynum = 2
        propertystring = "This is my test"
        evgen.event.PropertyNum = propertynum
        evgen.event.PropertyString = propertystring
        evgen.event.SourceNode = o.nodeid
        tid = datetime.utcnow()
        msg = "this is my msg "
        evgen.trigger(tid, msg)

        ev = myhandler.future.result(10)
        self.assertIsNot(ev, None)  # we did not receive event
        self.assertEqual(ev.EventType, etype.nodeid)
        self.assertEqual(ev.Severity, 1)
        self.assertEqual(ev.SourceName, 'MyObject')
        self.assertEqual(ev.SourceNode, o.nodeid)
        self.assertEqual(ev.Message.Text, msg)
        self.assertEqual(ev.Time, tid)
        self.assertEqual(ev.PropertyNum, propertynum)
        self.assertEqual(ev.PropertyString, propertystring)

        # time.sleep(0.1)
        sub.unsubscribe(handle)
        sub.delete()

    def test_several_different_events(self):
        objects = self.srv.get_objects_node()
        o = objects.add_object(3, 'MyObject')

        etype1 = self.srv.create_custom_event_type(2, 'MyEvent1', ua.ObjectIds.BaseEventType, [('PropertyNum', ua.VariantType.Float), ('PropertyString', ua.VariantType.String)])
        evgen1 = self.srv.get_event_generator(etype1, o)

        etype2 = self.srv.create_custom_event_type(2, 'MyEvent2', ua.ObjectIds.BaseEventType, [('PropertyNum', ua.VariantType.Float), ('PropertyString', ua.VariantType.String)])
        evgen2 = self.srv.get_event_generator(etype2, o)

        myhandler = MySubHandler2()
        sub = self.opc.create_subscription(100, myhandler)
        handle = sub.subscribe_events(o, etype1)

        propertynum1 = 1
        propertystring1 = "This is my test 1"
        evgen1.event.PropertyNum = propertynum1
        evgen1.event.PropertyString = propertystring1

        propertynum2 = 2
        propertystring2 = "This is my test 2"
        evgen2.event.PropertyNum = propertynum2
        evgen2.event.PropertyString = propertystring2

        for i in range(3):
            evgen1.trigger()
            evgen2.trigger()
        time.sleep(1)

        self.assertEqual(len(myhandler.results), 3)
        ev = myhandler.results[-1]
        self.assertEqual(ev.EventType, etype1.nodeid)

        handle = sub.subscribe_events(o, etype2)
        for i in range(4):
            evgen1.trigger()
            evgen2.trigger()
        time.sleep(1)


        ev1s = [ev for ev in myhandler.results if ev.EventType == etype1.nodeid]
        ev2s = [ev for ev in myhandler.results if ev.EventType == etype2.nodeid]

        self.assertEqual(len(myhandler.results), 11)
        self.assertEqual(len(ev2s), 4)
        self.assertEqual(len(ev1s), 7)

        sub.unsubscribe(handle)
        sub.delete()

    def test_several_different_events_2(self):
        objects = self.srv.get_objects_node()
        o = objects.add_object(3, 'MyObject')

        etype1 = self.srv.create_custom_event_type(2, 'MyEvent1', ua.ObjectIds.BaseEventType, [('PropertyNum', ua.VariantType.Float), ('PropertyString', ua.VariantType.String)])
        evgen1 = self.srv.get_event_generator(etype1, o)

        etype2 = self.srv.create_custom_event_type(2, 'MyEvent2', ua.ObjectIds.BaseEventType, [('PropertyNum2', ua.VariantType.Float), ('PropertyString', ua.VariantType.String)])
        evgen2 = self.srv.get_event_generator(etype2, o)

        etype3 = self.srv.create_custom_event_type(2, 'MyEvent3', ua.ObjectIds.BaseEventType, [('PropertyNum3', ua.VariantType.Float), ('PropertyString', ua.VariantType.String)])
        evgen3 = self.srv.get_event_generator(etype3, o)

        myhandler = MySubHandler2()
        sub = self.opc.create_subscription(100, myhandler)
        handle = sub.subscribe_events(o, [etype1, etype3])

        propertynum1 = 1
        propertystring1 = "This is my test 1"
        evgen1.event.PropertyNum = propertynum1
        evgen1.event.PropertyString = propertystring1

        propertynum2 = 2
        propertystring2 = "This is my test 2"
        evgen2.event.PropertyNum2 = propertynum2
        evgen2.event.PropertyString = propertystring2

        propertynum3 = 3
        propertystring3 = "This is my test 3"
        evgen3.event.PropertyNum3 = propertynum3
        evgen3.event.PropertyString = propertystring2

        for i in range(3):
            evgen1.trigger()
            evgen2.trigger()
            evgen3.trigger()
        evgen3.event.PropertyNum3 = 9999
        evgen3.trigger()
        time.sleep(1)

        ev1s = [ev for ev in myhandler.results if ev.EventType == etype1.nodeid]
        ev2s = [ev for ev in myhandler.results if ev.EventType == etype2.nodeid]
        ev3s = [ev for ev in myhandler.results if ev.EventType == etype3.nodeid]

        self.assertEqual(len(myhandler.results), 7)
        self.assertEqual(len(ev1s), 3)
        self.assertEqual(len(ev2s), 0)
        self.assertEqual(len(ev3s), 4)
        self.assertEqual(ev1s[0].PropertyNum, propertynum1)
        self.assertEqual(ev3s[0].PropertyNum3, propertynum3)
        self.assertEqual(ev3s[-1].PropertyNum3, 9999)
        self.assertEqual(ev1s[0].PropertyNum3, None)

        sub.unsubscribe(handle)
        sub.delete()

    def test_subscription_reconciliate(self):
        myhandler = SubHandler()
        sub = self.opc.create_subscription(100, myhandler)
        sub.has_unknown_handlers = True
        get_mi_response = [[1, 2, 3, 4], [301, 201, 202, 302]]
        mi_map = {301: myhandler, 302: myhandler}
        sub.unsubscribe = mock.Mock()
        with mock.patch.dict(sub._monitoreditems_map, mi_map):
            mi_del = sub.reconciliate(get_mi_response)
            mi_srv_only = set(get_mi_response[1]) - set(mi_map.keys())
            for mi_handler in (2, 3):
                self.assertIn(mock.call(mi_handler), sub.unsubscribe.call_args_list)
            for mi_handler in (1, 4):
                self.assertNotIn(mock.call(mi_handler), sub.unsubscribe.call_args_list)
            self.assertEqual(mi_del, len(mi_srv_only))
            self.assertFalse(sub.has_unknown_handlers)

class CustomInternalSession(InternalSession):
    TIMEOUT = 4

    def create_subscription(self, *args, **kwargs):
        time.sleep(self.TIMEOUT)
        return super(CustomInternalSession, self).create_subscription(*args, **kwargs)


class SubscriptionTestCustomServer(unittest.TestCase):
    PORT_NUM2 = 48511  # Needs to be different than num_port1 in tests_client.py

    @classmethod
    def setUpClass(cls):
        iserver = InternalServer(session_cls=CustomInternalSession)
        cls.srv = Server(iserver=iserver)
        iserver._parent = cls.srv
        cls.srv.set_endpoint('opc.tcp://127.0.0.1:{0:d}'.format(cls.PORT_NUM2))
        cls.srv.start()

        # start admin client
        # short timeout since we want it to fail
        cls.clt = Client('opc.tcp://admin@127.0.0.1:{0:d}'.format(cls.PORT_NUM2), timeout=1)
        cls.clt.connect()
        cls.opc = cls.clt

    @classmethod
    def tearDownClass(cls):
        cls.clt.disconnect()
        cls.srv.stop()

    def test_subscription_timeout_deadlock(self):
        self.assertRaises(
            TimeoutError,
            self.opc.create_subscription,
            100,
            MySubHandler()
        )
        time.sleep(5)
        # There is no better way to test if we actually did complete
        subs = self.opc.uaclient.registered_subscriptions()
        self.assertEqual(1, len(subs))
        self.assertTrue(subs[0].is_ready())



import time
from datetime import datetime, timedelta
import unittest

from opcua import Client
from opcua import Server
from opcua import ua

from opcua.server.history_sql import HistorySQLite
from opcua.server.history import HistoryDict

from tests_common import CommonTests, add_server_methods

from opcua.common.events import get_event_properties_from_type_node as get_props

port_num1 = 48530
port_num2 = 48530


class HistoryCommon(object):
    srv = Server
    clt = Client

    @classmethod
    def start_server_and_client(cls):
        cls.srv = Server()
        cls.srv.set_endpoint('opc.tcp://127.0.0.1:{0:d}'.format(port_num1))
        cls.srv.start()

        cls.clt = Client('opc.tcp://127.0.0.1:{0:d}'.format(port_num1))
        cls.clt.connect()

    @classmethod
    def stop_server_and_client(cls):
        cls.clt.disconnect()
        cls.srv.stop()

    @classmethod
    def create_var(cls):
        o = cls.srv.get_objects_node()
        cls.values = [i for i in range(20)]
        cls.var = o.add_variable(3, "history_var", 0)
        cls.srv.historize_node_data_change(cls.var, period=None, count=0)
        for i in cls.values:
            cls.var.set_value(i)
        time.sleep(1)

    # no start and no end is not defined by spec, return reverse order
    def test_history_var_read_one(self):
        # Spec says that at least two parameters should be provided, so
        # this one is out of spec
        res = self.var.read_raw_history(None, None, 1)
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0].Value.Value, self.values[-1])

    # no start and no end is not defined by spec, return reverse order
    def test_history_var_read_none(self):
        res = self.var.read_raw_history(None, None, 0)
        self.assertEqual(len(res), 20)
        self.assertEqual(res[0].Value.Value, self.values[-1])
        self.assertEqual(res[-1].Value.Value, self.values[0])

    # no start and no end is not defined by spec, return reverse order
    def test_history_var_read_last_3(self):
        res = self.var.read_raw_history(None, None, 3)
        self.assertEqual(len(res), 3)
        self.assertEqual(res[-1].Value.Value, self.values[-3])
        self.assertEqual(res[0].Value.Value, self.values[-1])

    # no start and no end is not defined by spec, return reverse order
    def test_history_var_read_all2(self):
        res = self.var.read_raw_history(None, None, 9999)
        self.assertEqual(len(res), 20)
        self.assertEqual(res[-1].Value.Value, self.values[0])
        self.assertEqual(res[0].Value.Value, self.values[-1])

    # only has end time, should return reverse order
    def test_history_var_read_2_with_end(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)

        res = self.var.read_raw_history(None, now, 2)
        self.assertEqual(len(res), 2)
        self.assertEqual(res[-1].Value.Value, self.values[-2])

    # both start and endtime, return from start to end
    def test_history_var_read_all(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)

        res = self.var.read_raw_history(old, now, 0)
        self.assertEqual(len(res), 20)
        self.assertEqual(res[-1].Value.Value, self.values[-1])
        self.assertEqual(res[0].Value.Value, self.values[0])

    def test_history_var_read_5_in_timeframe(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)

        res = self.var.read_raw_history(old, now, 5)
        self.assertEqual(len(res), 5)
        self.assertEqual(res[-1].Value.Value, self.values[4])
        self.assertEqual(res[0].Value.Value, self.values[0])

    # start time greater than end time, should return reverse order
    def test_history_var_read_5_in_timeframe_start_greater_than_end(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)

        res = self.var.read_raw_history(now, old, 5)
        self.assertEqual(len(res), 5)
        self.assertEqual(res[-1].Value.Value, self.values[-5])
        self.assertEqual(res[0].Value.Value, self.values[-1])

    # only start return original order
    def test_history_var_read_6_with_start(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)
        res = self.var.read_raw_history(old, None, 6)
        self.assertEqual(len(res), 6)
        self.assertEqual(res[-1].Value.Value, self.values[5])
        self.assertEqual(res[0].Value.Value, self.values[0])

    # only start return original order
    def test_history_var_read_all_with_start(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)
        res = self.var.read_raw_history(old, None, 0)
        self.assertEqual(len(res), 20)
        self.assertEqual(res[-1].Value.Value, self.values[-1])
        self.assertEqual(res[0].Value.Value, self.values[0])

    # only end return reversed order
    def test_history_var_read_all_with_end(self):
        end = datetime.utcnow() + timedelta(days=6)
        res = self.var.read_raw_history(None, end, 0)
        self.assertEqual(len(res), 20)
        self.assertEqual(res[-1].Value.Value, self.values[0])
        self.assertEqual(res[0].Value.Value, self.values[-1])

    # only end return reversed order
    def test_history_var_read_3_with_end(self):
        end = datetime.utcnow() + timedelta(days=6)
        res = self.var.read_raw_history(None, end, 3)
        self.assertEqual(len(res), 3)
        self.assertEqual(res[2].Value.Value, self.values[-3])
        self.assertEqual(res[0].Value.Value, self.values[-1])


class TestHistoryEvents(object):

    @classmethod
    def create_srv_events(cls):
        cls.ev_values = [i for i in range(20)]
        cls.srvevgen = cls.srv.get_event_generator()

        cls.srv_node = cls.srv.get_node(ua.ObjectIds.Server)
        cls.srv.historize_node_event(cls.srv_node, period=None)

        for i in cls.ev_values:
            cls.srvevgen.event.Severity = cls.ev_values[i]
            cls.srvevgen.trigger(message="test message")
            time.sleep(.1)
        time.sleep(2)

    # only has end time, should return reverse order
    def test_history_ev_read_2_with_end(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)

        res = self.srv_node.read_event_history(None, now, 2)
        self.assertEqual(len(res), 2)
        self.assertEqual(res[-1].Severity, self.ev_values[-2])

    # both start and end time, return from start to end
    def test_history_ev_read_all(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)

        res = self.srv_node.read_event_history(old, now, 0)
        self.assertEqual(len(res), 20)
        self.assertEqual(res[-1].Severity, self.ev_values[-1])
        self.assertEqual(res[0].Severity, self.ev_values[0])

    def test_history_ev_read_5_in_timeframe(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)

        res = self.srv_node.read_event_history(old, now, 5)
        self.assertEqual(len(res), 5)
        self.assertEqual(res[-1].Severity, self.ev_values[4])
        self.assertEqual(res[0].Severity, self.ev_values[0])

    # start time greater than end time, should return reverse order
    def test_history_ev_read_5_in_timeframe_start_greater_than_end(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)

        res = self.srv_node.read_event_history(now, old, 5)
        self.assertEqual(len(res), 5)
        self.assertEqual(res[-1].Severity, self.ev_values[-5])
        self.assertEqual(res[0].Severity, self.ev_values[-1])

    # only start return original order
    def test_history_ev_read_6_with_start(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)
        res = self.srv_node.read_event_history(old, None, 6)
        self.assertEqual(len(res), 6)
        self.assertEqual(res[-1].Severity, self.ev_values[5])
        self.assertEqual(res[0].Severity, self.ev_values[0])

    # only start return original order
    def test_history_ev_read_all_with_start(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)
        res = self.srv_node.read_event_history(old, None, 0)
        self.assertEqual(len(res), 20)
        self.assertEqual(res[-1].Severity, self.ev_values[-1])
        self.assertEqual(res[0].Severity, self.ev_values[0])

    # only end return reversed order
    def test_history_ev_read_all_with_end(self):
        end = datetime.utcnow() + timedelta(days=6)
        res = self.srv_node.read_event_history(None, end, 0)
        self.assertEqual(len(res), 20)
        self.assertEqual(res[-1].Severity, self.ev_values[0])
        self.assertEqual(res[0].Severity, self.ev_values[-1])

    # only end return reversed order
    def test_history_ev_read_3_with_end(self):
        end = datetime.utcnow() + timedelta(days=6)
        res = self.srv_node.read_event_history(None, end, 3)
        self.assertEqual(len(res), 3)
        self.assertEqual(res[2].Severity, self.ev_values[-3])
        self.assertEqual(res[0].Severity, self.ev_values[-1])

    # reverse event filter select clauses and test that results match the filter order
    def test_history_ev_read_all_filter_order_reversed(self):
        now = datetime.utcnow()
        old = now - timedelta(days=6)
        res = self.srv_node.read_event_history(old, None, 0)
        self.assertEqual(len(res), 20)
        self.assertEqual(res[-1].Severity, self.ev_values[-1])
        self.assertEqual(res[0].Severity, self.ev_values[0])


class TestHistoryLimitsCommon(unittest.TestCase):
    id = ua.NodeId(123)

    def setUp(self):
        self.history = self.createHistoryInstance()

    def createHistoryInstance(self):
        assert(False)

    def resultCount(self):
        results, cont = self.history.read_node_history(self.id, None, None, None)
        return len(results)

    def addValue(self, age):
        value = ua.DataValue()
        value.SourceTimestamp = datetime.utcnow() - timedelta(hours = age)
        self.history.save_node_value(self.id, value)

    def test_count_limit(self):
        self.history.new_historized_node(self.id, period=None, count=3)
        self.assertEqual(self.resultCount(), 0)
        self.addValue(5)
        self.assertEqual(self.resultCount(), 1)
        self.addValue(4)
        self.assertEqual(self.resultCount(), 2)
        self.addValue(3)
        self.assertEqual(self.resultCount(), 3)
        self.addValue(2)
        self.assertEqual(self.resultCount(), 3)
        self.addValue(1)
        self.assertEqual(self.resultCount(), 3)

    def test_period_limit(self):
        self.history.new_historized_node(self.id, period=timedelta(hours=3))
        self.assertEqual(self.resultCount(), 0)
        self.addValue(5)
        self.assertEqual(self.resultCount(), 0)
        self.addValue(4)
        self.assertEqual(self.resultCount(), 0)
        self.addValue(2)
        self.assertEqual(self.resultCount(), 1)
        self.addValue(1)
        self.assertEqual(self.resultCount(), 2)
        self.addValue(0)
        self.assertEqual(self.resultCount(), 3)

    def test_combined_limit(self):
        self.history.new_historized_node(self.id, period=timedelta(hours=3), count=2)
        self.assertEqual(self.resultCount(), 0)
        self.addValue(5)
        self.assertEqual(self.resultCount(), 0)
        self.addValue(4)
        self.assertEqual(self.resultCount(), 0)
        self.addValue(2)
        self.assertEqual(self.resultCount(), 1)
        self.addValue(1)
        self.assertEqual(self.resultCount(), 2)
        self.addValue(0)
        self.assertEqual(self.resultCount(), 2)


class TestHistoryLimits(TestHistoryLimitsCommon):
    def createHistoryInstance(self):
        return HistoryDict()


class TestHistorySQLLimits(TestHistoryLimitsCommon):
    def createHistoryInstance(self):
        return HistorySQLite(":memory:")


class TestHistory(unittest.TestCase, HistoryCommon, TestHistoryEvents):

    @classmethod
    def setUpClass(cls):
        cls.start_server_and_client()
        cls.create_var()
        cls.create_srv_events()

    @classmethod
    def tearDownClass(cls):
        cls.stop_server_and_client()


class TestHistorySQL(unittest.TestCase, HistoryCommon, TestHistoryEvents):
    @classmethod
    def setUpClass(cls):
        cls.start_server_and_client()
        cls.srv.iserver.history_manager.set_storage(HistorySQLite(":memory:"))
        cls.create_var()
        cls.create_srv_events()

    @classmethod
    def tearDownClass(cls):
        cls.stop_server_and_client()



import unittest
import logging
import xml.etree.ElementTree as Et
from opcua import ua, Server
import opcua.common.type_dictionary_buider
from opcua.common.type_dictionary_buider import OPCTypeDictionaryBuilder, DataTypeDictionaryBuilder
from opcua.common.type_dictionary_buider import get_ua_class, StructNode

port_num = 48540
idx_name = 'http://test.freeopcua.github.io'


def to_camel_case(name):
    func = getattr(opcua.common.type_dictionary_buider, '_to_camel_case')
    return func(name)


def reference_generator(source_id, target_id, reference_type, is_forward=True):
    func = getattr(opcua.common.type_dictionary_buider, '_reference_generator')
    return func(source_id, target_id, reference_type, is_forward)


def set_up_test_tree():
    ext_head_attributes = {'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance', 'xmlns:tns': idx_name,
                           'DefaultByteOrder': 'LittleEndian', 'xmlns:opc': 'http://opcfoundation.org/BinarySchema/',
                           'xmlns:ua': 'http://opcfoundation.org/UA/', 'TargetNamespace': idx_name}

    test_etree = Et.ElementTree(Et.Element('opc:TypeDictionary', ext_head_attributes))
    name_space = Et.SubElement(test_etree.getroot(), 'opc:Import')
    name_space.attrib['Namespace'] = 'http://opcfoundation.org/UA/'
    return test_etree


class TypeDictionaryBuilderTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.srv = Server()
        cls.srv.set_endpoint('opc.tcp://127.0.0.1:{0:d}'.format(port_num))
        cls.idx = cls.srv.register_namespace(idx_name)
        cls.srv.start()

    @classmethod
    def tearDownClass(cls):
        cls.srv.stop()

    def setUp(self):
        self.test_etree = set_up_test_tree()
        self.opc_type_builder = OPCTypeDictionaryBuilder(idx_name)
        self.dict_builder = DataTypeDictionaryBuilder(self.srv, self.idx, idx_name, 'TestDict')

    def tearDown(self):
        return

    def test_camel_case_1(self):
        case = 'TurtleActionlibShapeActionFeedback'
        result = to_camel_case('turtle_actionlib/ShapeActionFeedback')
        self.assertEquals(result, case)

    def test_camel_case_2(self):
        case = 'HelloWorldFffD'
        result = to_camel_case('Hello#world+fff_**?&&d')
        self.assertEquals(result, case)

    def test_opc_type_dict_process_type_opc(self):
        case = 'opc:Boolean'
        result = getattr(self.opc_type_builder, '_process_type')('Boolean')
        self.assertEquals(result, case)

    def test_opc_type_dict_process_type_tns(self):
        case = 'tns:CustomizedStruct'
        result = getattr(self.opc_type_builder, '_process_type')('CustomizedStruct')
        self.assertEquals(result, case)

    def test_opc_type_dict_append_struct_1(self):
        case = {'BaseType': 'ua:ExtensionObject',
                'Name': 'CustomizedStruct'}
        result = self.opc_type_builder.append_struct('CustomizedStruct')
        self.assertEquals(result.attrib, case)

    def DISABLED_test_opc_type_dict_append_struct_2(self):
        case = {'BaseType': 'ua:ExtensionObject',
                'Name': 'CustomizedStruct'}
        result = self.opc_type_builder.append_struct('customized_#?+`struct')
        self.assertEquals(result.attrib, case)

    def test_opc_type_dict_add_field_1(self):
        structure_name = 'CustomizedStruct'
        self.opc_type_builder.append_struct(structure_name)
        self.opc_type_builder.add_field('id', ua.VariantType.Boolean, structure_name)
        case = {'TypeName': 'opc:Boolean',
                'Name': 'id'}
        struct_dict = getattr(self.opc_type_builder, '_structs_dict')
        result = list(struct_dict[structure_name])[0]
        self.assertEquals(result.attrib, case)

    def test_opc_type_dict_add_field_2(self):
        structure_name = 'CustomizedStruct'
        self.opc_type_builder.append_struct(structure_name)
        self.opc_type_builder.add_field('id', 'Boolean', structure_name)
        case = {'TypeName': 'opc:Boolean',
                'Name': 'id'}
        struct_dict = getattr(self.opc_type_builder, '_structs_dict')
        result = list(struct_dict[structure_name])[0]
        self.assertEquals(result.attrib, case)

    def test_opc_type_dict_add_field_3(self):
        structure_name = 'CustomizedStruct'
        self.opc_type_builder.append_struct(structure_name)
        self.opc_type_builder.add_field('id', ua.VariantType.Boolean, structure_name, is_array=True)
        case = [{'TypeName': 'opc:Int32',
                'Name': 'NoOfid'},
                {'TypeName': 'opc:Boolean',
                 'LengthField': 'NoOfid',
                 'Name': 'id'}]
        struct_dict = getattr(self.opc_type_builder, '_structs_dict')
        result = [item.attrib for item in list(struct_dict[structure_name])]
        self.assertTrue(len(case), len(result))
        for item in case:
            self.assertTrue(item in result)

    def test_opc_type_dict_get_dict_value(self):
        structure_name = 'CustomizedStruct'
        self.opc_type_builder.append_struct(structure_name)
        # external tree operation
        appended_struct = Et.SubElement(self.test_etree.getroot(), 'opc:StructuredType')
        appended_struct.attrib['BaseType'] = 'ua:ExtensionObject'
        appended_struct.attrib['Name'] = to_camel_case(structure_name)

        self.opc_type_builder.add_field('id', ua.VariantType.Boolean, structure_name)
        # external tree operation
        field = Et.SubElement(appended_struct, 'opc:Field')
        field.attrib['Name'] = 'id'
        field.attrib['TypeName'] = 'opc:Boolean'
        case = Et.tostring(self.test_etree.getroot(), encoding='utf-8').decode("utf-8").replace(' ', '')
        result = self.opc_type_builder.get_dict_value().decode("utf-8").replace(' ', '').replace('\n', '')
        self.assertEqual(result, case)

    def test_reference_generator_1(self):
        id1 = ua.NodeId(1, namespaceidx=2, nodeidtype=ua.NodeIdType.Numeric)
        id2 = ua.NodeId(2, namespaceidx=2, nodeidtype=ua.NodeIdType.Numeric)
        ref = ua.NodeId(ua.ObjectIds.HasEncoding, 0)
        result = reference_generator(id1, id2, ref)
        self.assertTrue(result.IsForward)
        self.assertEqual(result.ReferenceTypeId, ref)
        self.assertEqual(result.SourceNodeId, id1)
        self.assertEqual(result.TargetNodeClass, ua.NodeClass.DataType)
        self.assertEqual(result.TargetNodeId, id2)

    def test_reference_generator_2(self):
        id1 = ua.NodeId(1, namespaceidx=2, nodeidtype=ua.NodeIdType.Numeric)
        id2 = ua.NodeId(2, namespaceidx=2, nodeidtype=ua.NodeIdType.Numeric)
        ref = ua.NodeId(ua.ObjectIds.HasEncoding, 0)
        result = reference_generator(id1, id2, ref, False)
        self.assertFalse(result.IsForward)
        self.assertEqual(result.ReferenceTypeId, ref)
        self.assertEqual(result.SourceNodeId, id1)
        self.assertEqual(result.TargetNodeClass, ua.NodeClass.DataType)
        self.assertEqual(result.TargetNodeId, id2)

    def test_data_type_dict_general(self):
        self.assertIsNotNone(self.dict_builder.dict_id)
        self.assertIsNotNone(getattr(self.dict_builder, '_type_dictionary'))

    def test_data_type_dict_add_dictionary(self):
        add_dictionary = getattr(self.dict_builder, '_add_dictionary')
        dict_name = 'TestDict'
        dict_node = self.srv.get_node(add_dictionary(dict_name))
        self.assertEqual(dict_node.get_browse_name(), ua.QualifiedName(dict_name, self.idx))
        self.assertEqual(dict_node.get_node_class(), ua.NodeClass.Variable)
        self.assertEqual(dict_node.get_parent().nodeid, ua.NodeId(ua.ObjectIds.OPCBinarySchema_TypeSystem, 0))
        self.assertEqual(ua.NodeId(ua.ObjectIds.HasComponent, 0),
                         dict_node.get_references(refs=ua.ObjectIds.HasComponent)[0].ReferenceTypeId)
        self.assertEqual(dict_node.get_type_definition(), ua.NodeId(ua.ObjectIds.DataTypeDictionaryType, 0))
        self.assertEqual(dict_node.get_display_name(), ua.LocalizedText(dict_name))
        self.assertEqual(dict_node.get_data_type(), ua.NodeId(ua.ObjectIds.ByteString))
        self.assertEqual(dict_node.get_value_rank(), -1)

    def test_data_type_dict_create_data_type(self):
        type_name = 'CustomizedStruct2'
        created_type = self.dict_builder.create_data_type(type_name)
        self.assertTrue(isinstance(created_type, StructNode))
        # Test data type node
        type_node = self.srv.get_node(created_type.data_type)
        self.assertEqual(type_node.get_browse_name(), ua.QualifiedName(type_name, self.idx))
        self.assertEqual(type_node.get_node_class(), ua.NodeClass.DataType)
        self.assertEqual(type_node.get_parent().nodeid, ua.NodeId(ua.ObjectIds.Structure, 0))
        self.assertEqual(ua.NodeId(ua.ObjectIds.HasSubtype, 0),
                         type_node.get_references(refs=ua.ObjectIds.HasSubtype)[0].ReferenceTypeId)
        self.assertEqual(type_node.get_display_name(), ua.LocalizedText(type_name))

        # Test description node
        desc_node = self.srv.get_node(self.dict_builder.dict_id).get_child("2:{}".format(type_name))
        self.assertEqual(desc_node.get_browse_name(), ua.QualifiedName(type_name, self.idx))
        self.assertEqual(desc_node.get_node_class(), ua.NodeClass.Variable)
        self.assertEqual(desc_node.get_parent().nodeid, self.dict_builder.dict_id)
        self.assertEqual(ua.NodeId(ua.ObjectIds.HasComponent, 0),
                         desc_node.get_references(refs=ua.ObjectIds.HasComponent)[0].ReferenceTypeId)
        self.assertEqual(desc_node.get_type_definition(), ua.NodeId(ua.ObjectIds.DataTypeDescriptionType, 0))

        self.assertEqual(desc_node.get_display_name(), ua.LocalizedText(type_name))
        self.assertEqual(desc_node.get_data_type(), ua.NodeId(ua.ObjectIds.String))
        self.assertEqual(desc_node.get_value(), type_name)
        self.assertEqual(desc_node.get_value_rank(), -1)

        # Test object node
        obj_node = type_node.get_children(refs=ua.ObjectIds.HasEncoding)[0]
        self.assertEqual(obj_node.get_browse_name(), ua.QualifiedName('Default Binary', 0))
        self.assertEqual(obj_node.get_node_class(), ua.NodeClass.Object)
        self.assertEqual(obj_node.get_references(refs=ua.ObjectIds.HasEncoding)[0].NodeId, type_node.nodeid)
        self.assertEqual(ua.NodeId(ua.ObjectIds.HasEncoding, 0),
                         obj_node.get_references(refs=ua.ObjectIds.HasEncoding)[0].ReferenceTypeId)
        self.assertEqual(obj_node.get_type_definition(), ua.NodeId(ua.ObjectIds.DataTypeEncodingType, 0))
        self.assertEqual(obj_node.get_display_name(), ua.LocalizedText('Default Binary'))
        self.assertEqual(len(obj_node.get_event_notifier()), 0)

        # Test links, three were tested above
        struct_node = self.srv.get_node(ua.NodeId(ua.ObjectIds.Structure, 0))
        struct_children = struct_node.get_children(refs=ua.ObjectIds.HasSubtype)
        self.assertTrue(type_node in struct_children)
        dict_node = self.srv.get_node(self.dict_builder.dict_id)
        dict_children = dict_node.get_children(refs=ua.ObjectIds.HasComponent)
        self.assertTrue(desc_node in dict_children)
        self.assertTrue(obj_node in type_node.get_children(ua.ObjectIds.HasEncoding))
        self.assertTrue(desc_node in obj_node.get_children(refs=ua.ObjectIds.HasDescription))
        self.assertEqual(obj_node.nodeid, desc_node.get_references(refs=ua.ObjectIds.HasDescription,
                                                                   direction=ua.BrowseDirection.Inverse)[0].NodeId)

    def test_data_type_dict_set_dict_byte_string(self):
        structure_name = 'CustomizedStruct'
        self.dict_builder.create_data_type(structure_name)
        self.dict_builder.add_field('id', ua.VariantType.Int32, structure_name)
        self.dict_builder.set_dict_byte_string()
        # external tree operation
        appended_struct = Et.SubElement(self.test_etree.getroot(), 'opc:StructuredType')
        appended_struct.attrib['BaseType'] = 'ua:ExtensionObject'
        appended_struct.attrib['Name'] = to_camel_case(structure_name)

        # external tree operation
        field = Et.SubElement(appended_struct, 'opc:Field')
        field.attrib['Name'] = 'id'
        field.attrib['TypeName'] = 'opc:Int32'
        case = Et.tostring(self.test_etree.getroot(), encoding='utf-8').decode("utf-8").replace(' ', '')
        result_string = self.srv.get_node(self.dict_builder.dict_id).get_value().decode("utf-8")
        result = result_string.replace(' ', '').replace('\n', '')
        self.assertEqual(result, case)

    def test_data_type_dict_add_field_1(self):
        struct_name = 'CustomizedStruct'
        self.dict_builder.create_data_type(struct_name)
        self.dict_builder.add_field('id', ua.VariantType.Int32, struct_name)
        self.dict_builder.set_dict_byte_string()
        self.srv.load_type_definitions()
        struct = get_ua_class(struct_name)
        self.assertEqual(struct.ua_types[0][0], 'id')
        self.assertEqual(struct.ua_types[0][1], 'Int32')
        struct_instance = struct()
        self.assertEqual(struct_instance.id, 0)

    def test_data_type_dict_add_field_2(self):
        struct_name = 'AnotherCustomizedStruct'
        self.dict_builder.create_data_type(struct_name)
        self.dict_builder.add_field('id', ua.VariantType.Int32, struct_name, is_array=True)
        self.dict_builder.set_dict_byte_string()
        self.srv.load_type_definitions()
        struct = get_ua_class(struct_name)
        self.assertEqual(struct.ua_types[0][0], 'id')
        self.assertEqual(struct.ua_types[0][1], 'ListOfInt32')
        struct_instance = struct()
        self.assertTrue(isinstance(struct_instance.id, list))

    def test_struct_node_general(self):
        struct_name = 'CustomizedStruct'
        struct_node = self.dict_builder.create_data_type(struct_name)
        self.assertEqual(getattr(struct_node, '_type_dict'), self.dict_builder)
        self.assertTrue(isinstance(struct_node.data_type, ua.NodeId))
        self.assertEqual(struct_node.name, struct_name)

    def test_struct_node_add_field(self):
        struct_name = 'CustomizedStruct'
        struct_node = self.dict_builder.create_data_type(struct_name)
        struct_node.add_field('id', ua.VariantType.Int32)
        self.dict_builder.set_dict_byte_string()
        self.srv.load_type_definitions()
        struct = get_ua_class(struct_name)
        self.assertEqual(struct.ua_types[0][0], 'id')
        self.assertEqual(struct.ua_types[0][1], 'Int32')
        struct_instance = struct()
        self.assertEqual(struct_instance.id, 0)

    def test_get_ua_class_1(self):
        struct_name = 'CustomizedStruct'
        struct_node = self.dict_builder.create_data_type(struct_name)
        struct_node.add_field('id', ua.VariantType.Int32)
        self.dict_builder.set_dict_byte_string()
        self.srv.load_type_definitions()
        try:
            self.assertIsNotNone(get_ua_class(struct_name))
        except AttributeError:
            pass

    def DISABLED_test_get_ua_class_2(self):
        struct_name = '*c*u_stom-ized&Stru#ct'
        struct_node = self.dict_builder.create_data_type(struct_name)
        struct_node.add_field('id', ua.VariantType.Int32)
        self.dict_builder.set_dict_byte_string()
        self.srv.load_type_definitions()
        try:
            self.assertIsNotNone(get_ua_class(struct_name))
        except AttributeError:
            pass

    def test_functional_basic(self):
        basic_struct_name = 'basic_structure'
        basic_struct = self.dict_builder.create_data_type(basic_struct_name)
        basic_struct.add_field('ID', ua.VariantType.Int32)
        basic_struct.add_field('Gender', ua.VariantType.Boolean)
        basic_struct.add_field('Comments', ua.VariantType.String)

        self.dict_builder.set_dict_byte_string()
        self.srv.load_type_definitions()

        basic_var = self.srv.nodes.objects.add_variable(ua.NodeId(namespaceidx=self.idx), 'BasicStruct',
                                                        ua.Variant(None, ua.VariantType.Null),
                                                        datatype=basic_struct.data_type)

        basic_msg = get_ua_class(basic_struct_name)()
        basic_msg.ID = 3
        basic_msg.Gender = True
        basic_msg.Comments = 'Test string'
        basic_var.set_value(basic_msg)

        basic_result = basic_var.get_value()
        self.assertEqual(basic_result, basic_msg)

    def test_functional_advance(self):
        basic_struct_name = 'base_structure'
        basic_struct = self.dict_builder.create_data_type(basic_struct_name)
        basic_struct.add_field('ID', ua.VariantType.Int32)
        basic_struct.add_field('Gender', ua.VariantType.Boolean)
        basic_struct.add_field('Comments', ua.VariantType.String)

        nested_struct_name = 'nested_structure'
        nested_struct = self.dict_builder.create_data_type(nested_struct_name)
        nested_struct.add_field('Name', ua.VariantType.String)
        nested_struct.add_field('Surname', ua.VariantType.String)
        nested_struct.add_field('Stuff', basic_struct)

        self.dict_builder.set_dict_byte_string()
        self.srv.load_type_definitions()

        basic_var = self.srv.nodes.objects.add_variable(ua.NodeId(namespaceidx=self.idx), 'BaseStruct',
                                                        ua.Variant(None, ua.VariantType.Null),
                                                        datatype=basic_struct.data_type)

        basic_msg = get_ua_class(basic_struct_name)()
        basic_msg.ID = 3
        basic_msg.Gender = True
        basic_msg.Comments = 'Test string'
        basic_var.set_value(basic_msg)

        nested_var = self.srv.nodes.objects.add_variable(ua.NodeId(namespaceidx=self.idx), 'NestedStruct',
                                                         ua.Variant(None, ua.VariantType.Null),
                                                         datatype=nested_struct.data_type)

        nested_msg = get_ua_class(nested_struct_name)()
        nested_msg.Stuff = basic_msg
        nested_msg.Name = 'Max'
        nested_msg.Surname = 'Karl'
        nested_var.set_value(nested_msg)

        basic_result = basic_var.get_value()
        self.assertEqual(basic_result, basic_msg)
        nested_result = nested_var.get_value()
        self.assertEqual(nested_result, nested_msg)

    def test_siemens_flavored_struct_name(self):
        struct_name = '"Siemens"."OPC-UA"."Identifiers_LookLikeThis"'
        struct_node = self.dict_builder.create_data_type(struct_name)
        struct_node.add_field('"a"."field"', ua.VariantType.Int32)
        self.dict_builder.set_dict_byte_string()
        self.srv.load_type_definitions()

if __name__ == '__main__':
    logging.basicConfig(level=logging.WARNING)
    unittest.main(verbosity=3)



import unittest
import logging
import sys
sys.path.insert(0, "..")
sys.path.insert(0, ".")


from tests_cmd_lines import TestCmdLines
from tests_server import TestServer, TestServerCaching, TestServerStartError
from tests_client import TestClient
from tests_subscriptions import SubscriptionTestCustomServer
from tests_standard_address_space import StandardAddressSpaceTests
from tests_unit import TestUnit, TestMaskEnum
from tests_history import TestHistory, TestHistorySQL, TestHistoryLimits, TestHistorySQLLimits
from tests_crypto_connect import TestCryptoConnect
from tests_uaerrors import TestUaErrors
from tests_custom_structures import TypeDictionaryBuilderTest


if __name__ == '__main__':
    logging.basicConfig(level=logging.WARNING)
    #l = logging.getLogger("opcua.server.internal_subscription")
    #l.setLevel(logging.DEBUG)
    #l = logging.getLogger("opcua.server.internal_server")
    #l.setLevel(logging.DEBUG)

    unittest.main(verbosity=3)



import inspect
import sys
import unittest
from datetime import datetime
from enum import EnumMeta

from opcua import ua
from opcua.common.structures import StructGenerator, Struct, EnumType
from opcua.ua.ua_binary import struct_to_binary, struct_from_binary


# An ExtensionObject that is decoded becomes a generated class of a custom name.
# The class below represents an example that includes OPC UA optional fields.
class ObjectWithOptionalFields(object):
    ua_switches = {
        'cavityId': ('BitEncoding0', 0),
        'description': ('BitEncoding0', 1),
        # This exists in xml but gets ignored because no matching switches.
        # 'Reserved1': ('ByteEncoding0', 2)
    }
    ua_types = [
        # The bits fields are not necessary because 'ua_switches' provides us with
        # the related bit we care about.
        # ('cavityIdSpecified', 'Bit'),
        # ('descriptionSpecified', 'Bit'),
        # ('Reserved1', 'Bit'),
        ('BitEncoding0', 'UInt32'),
        ('name', 'CharArray'),
        ('value', 'Double'),
        ('assignment', 'UInt32'),
        ('source', 'UInt32'),
        ('cavityId', 'UInt32'),
        ('id', 'CharArray'),
        ('description', 'CharArray'),
    ]

    def __str__(self):
        vals = [name + ": " + str(val) for name, val in self.__dict__.items()]
        return self.__class__.__name__ + "(" + ", ".join(vals) + ")"

    __repr__ = __str__

    def __init__(self):
        self.BitEncoding0 = 0x03
        self.name = 'SomeAmazingCycleParameter'
        self.value = 8
        self.assignment = 8
        self.source = 8
        self.cavityId = 5
        self.id = 'abcdefgh'
        self.description = 'BBA'


class CustomStructTestCase(unittest.TestCase):

    def setUp(self):
        self.identifier_count = 0

    def _generate_node_id(self):
        self.identifier_count += 1
        return f"ns=0;i={self.identifier_count}"

    @staticmethod
    def is_struct(obj):
        # TODO: putting this definition in the generated class would be better
        return hasattr(obj, 'ua_types')

    def assertCustomStructEqual(self, original, deserialized):
        if hasattr(original, 'ua_switches'):
            self.assertEqual(len(original.ua_switches), len(deserialized.ua_switches))
        self.assertEqual(len(original.ua_types), len(deserialized.ua_types))
        for field, _ in original.ua_types:
            field_obj = getattr(original, field)
            deserialized_obj = getattr(deserialized, field)
            if self.is_struct(field_obj):
                self.assertCustomStructEqual(field_obj, deserialized_obj)
            else:
                self.assertEqual(getattr(original, field), getattr(deserialized, field))

    def test_binary_struct_example(self):
        # Example test so that we can manually control the object that gets
        # generated and see how it gets serialized/deserialized
        original = ObjectWithOptionalFields()
        serialized = struct_to_binary(original)
        deserialized = struct_from_binary(ObjectWithOptionalFields, ua.utils.Buffer(serialized))
        self.assertEqual(len(original.ua_switches), len(deserialized.ua_switches))
        self.assertEqual(len(original.ua_types), len(deserialized.ua_types))
        for field, _ in original.ua_types:
            self.assertEqual(getattr(original, field), getattr(deserialized, field))

    def test_custom_struct_with_optional_fields(self):
        xmlpath = "custom_extension_with_optional_fields.xml"
        c = StructGenerator()
        c.make_model_from_file(xmlpath)
        for m in c.model:
            if type(m) in (Struct, EnumType):
                m.typeid = self._generate_node_id()
        c.save_to_file("custom_extension_with_optional_fields.py", register=True)
        import como_structures as s
        for name, obj in inspect.getmembers(sys.modules[s.__name__], predicate=inspect.isclass):
            if name.startswith('__') or obj in (datetime,) or isinstance(obj, EnumMeta):
                continue
            with self.subTest(name=name):
                original = obj()
                serialized = struct_to_binary(original)
                deserialized = struct_from_binary(obj, ua.utils.Buffer(serialized))
                self.assertCustomStructEqual(original, deserialized)



import unittest
import opcua.ua.uaerrors as uaerrors
from opcua.ua.uaerrors import UaStatusCodeError

class TestUaErrors(unittest.TestCase):
    status_code_bad_internal = 0x80020000
    status_code_unknown = "Definitely Not A Status Code"

    def setUp(self):
        self.direct = uaerrors.BadInternalError()
        self.indirect = UaStatusCodeError(self.status_code_bad_internal)
        self.unknown = UaStatusCodeError(self.status_code_unknown)

    def test_subclass_selection(self):
        self.assertIs(type(self.direct), uaerrors.BadInternalError)
        self.assertIs(type(self.indirect), uaerrors.BadInternalError)
        self.assertIs(type(self.unknown), UaStatusCodeError)

    def test_code(self):
        self.assertEqual(self.direct.code, self.status_code_bad_internal)
        self.assertEqual(self.indirect.code, self.status_code_bad_internal)
        self.assertEqual(self.unknown.code, self.status_code_unknown)

    def test_string_repr(self):
        self.assertIn("BadInternal", str(self.direct))
        self.assertIn("BadInternal", str(self.indirect))



import unittest

from opcua import Client
from opcua import Server
from opcua import ua
from opcua.client.ua_client import UASocketClient
from opcua.common.utils import SocketWrapper

from tests_subscriptions import SubscriptionTests, SubHandler
from tests_common import CommonTests, add_server_methods
from tests_xml import XmlTests

from tests_enum_struct import add_server_custom_enum_struct

try:
    from unittest import mock
except ImportError:
    import mock

port_num1 = 48510


class TestClient(unittest.TestCase, CommonTests, SubscriptionTests, XmlTests):

    '''
    Run common tests on client side
    Of course we need a server so we start also start a server
    Tests that can only be run on client side must be defined  in this class
    '''
    @classmethod
    def setUpClass(cls):
        # start our own server
        cls.srv = Server()
        cls.srv.set_endpoint('opc.tcp://127.0.0.1:{0:d}'.format(port_num1))
        add_server_methods(cls.srv)
        add_server_custom_enum_struct(cls.srv)
        cls.srv.start()

        # start admin client
        # long timeout since travis (automated testing) can be really slow
        cls.clt = Client('opc.tcp://admin@127.0.0.1:{0:d}'.format(port_num1), timeout=10)
        cls.clt.connect()
        cls.opc = cls.clt

        # start anonymous client
        cls.ro_clt = Client('opc.tcp://127.0.0.1:{0:d}'.format(port_num1))
        cls.ro_clt.connect()

    @classmethod
    def tearDownClass(cls):
        #stop our clients
        cls.ro_clt.disconnect()
        cls.clt.disconnect()
        # stop the server
        cls.srv.stop()

    def test_service_fault(self):
        request = ua.ReadRequest()
        request.TypeId = ua.FourByteNodeId(999)  # bad type!
        with self.assertRaises(ua.UaStatusCodeError):
            self.clt.uaclient._uasocket.send_request(request)

    def test_objects_anonymous(self):
        objects = self.ro_clt.get_objects_node()
        with self.assertRaises(ua.UaStatusCodeError):
            objects.set_attribute(ua.AttributeIds.WriteMask, ua.DataValue(999))
        with self.assertRaises(ua.UaStatusCodeError):
            f = objects.add_folder(3, 'MyFolder')

    def test_folder_anonymous(self):
        objects = self.clt.get_objects_node()
        f = objects.add_folder(3, 'MyFolderRO')
        f_ro = self.ro_clt.get_node(f.nodeid)
        self.assertEqual(f, f_ro)
        with self.assertRaises(ua.UaStatusCodeError):
            f2 = f_ro.add_folder(3, 'MyFolder2')

    def test_variable_anonymous(self):
        objects = self.clt.get_objects_node()
        v = objects.add_variable(3, 'MyROVariable', 6)
        v.set_value(4)  # this should work
        v_ro = self.ro_clt.get_node(v.nodeid)
        with self.assertRaises(ua.UaStatusCodeError):
            v_ro.set_value(2)
        self.assertEqual(v_ro.get_value(), 4)
        v.set_writable(True)
        v_ro.set_value(2)  # now it should work
        self.assertEqual(v_ro.get_value(), 2)
        v.set_writable(False)
        with self.assertRaises(ua.UaStatusCodeError):
            v_ro.set_value(9)
        self.assertEqual(v_ro.get_value(), 2)

    def test_multiple_read_and_write(self):
        objects = self.srv.get_objects_node()
        f = objects.add_folder(3, 'Multiple_read_write_test')
        v1 = f.add_variable(3, "a", 1)
        v1.set_writable()
        v2 = f.add_variable(3, "b", 2)
        v2.set_writable()
        v3 = f.add_variable(3, "c", 3)
        v3.set_writable()
        v_ro = f.add_variable(3, "ro", 3)

        vals = self.ro_clt.get_values([v1, v2, v3])
        self.assertEqual(vals, [1, 2, 3])
        self.ro_clt.set_values([v1, v2, v3], [4, 5, 6])
        vals = self.ro_clt.get_values([v1, v2, v3])
        self.assertEqual(vals, [4, 5, 6])
        with self.assertRaises(ua.uaerrors.BadUserAccessDenied):
            self.ro_clt.set_values([v1, v2, v_ro], [4, 5, 6])

    def test_context_manager(self):
        """ Context manager calls connect() and disconnect()
        """
        state = [0]
        def increment_state(self, *args, **kwargs):
            state[0] += 1

        # create client and replace instance methods with dummy methods
        client = Client('opc.tcp://dummy_address:10000')
        client.connect    = increment_state.__get__(client)
        client.disconnect = increment_state.__get__(client)

        assert state[0] == 0
        with client:
            # test if client connected
            self.assertEqual(state[0], 1)
        # test if client disconnected
        self.assertEqual(state[0], 2)

    def test_enumstrings_getvalue(self):
        ''' The real exception is server side, but is detected by using a client.
            Alldue the server trace is also visible on the console.
            The client only 'sees' an TimeoutError
        '''
        nenumstrings = self.clt.get_node(ua.ObjectIds.AxisScaleEnumeration_EnumStrings)
        with self.assertNotRaises(Exception):
            value = ua.Variant(nenumstrings.get_value())

    def test_uasocketclient_connect_disconnect(self):
        """Initialize, connect, and disconnect a UaSocketClient
        """
        uaclt = UASocketClient()
        uaclt.connect_socket('127.0.0.1', port_num1)
        self.assertTrue(uaclt._thread.is_alive())
        self.assertIsInstance(uaclt._socket, SocketWrapper)

        # disconnect_socket() should shut down the receiving thread
        uaclt.disconnect_socket()
        self.assertFalse(uaclt._thread.is_alive())

    def test_custom_enum_struct(self):
        self.ro_clt.load_type_definitions()
        ns = self.ro_clt.get_namespace_index('http://yourorganisation.org/struct_enum_example/')
        myvar = self.ro_clt.get_node(ua.NodeId(6009, ns))
        val = myvar.get_value()
        self.assertEqual(val.IntVal1, 242)
        self.assertEqual(val.EnumVal, ua.ExampleEnum.EnumVal2)

    @mock.patch("opcua.common.subscription.Subscription.reconciliate")
    @mock.patch("opcua.common.node.Node.call_method")
    def test_reconciliate(self, mock_call_method, mock_reconciliate):
        get_mi_response = [[1, 2], [201, 202]]
        mock_call_method.return_value = get_mi_response

        myhandler = SubHandler()
        sub = self.opc.create_subscription(100, myhandler)
        self.clt.reconciliate_subscription(sub)

        node_called_args = mock_call_method.call_args[0]
        self.assertEqual(
            node_called_args[0], ua.uatypes.QualifiedName("GetMonitoredItems")
        )
        self.assertEqual(
            node_called_args[1], ua.Variant(sub.subscription_id, ua.VariantType.UInt32)
        )

        reconciliate_called_args = mock_reconciliate.call_args[0][0]
        self.assertEqual(reconciliate_called_args, get_mi_response)



import unittest
import subprocess

from opcua import Server


port_num = 48530


class TestCmdLines(unittest.TestCase):

    '''
    Test command lines
    '''
    @classmethod
    def setUpClass(cls):
        cls.srv = Server()
        cls.srv_url = 'opc.tcp://127.0.0.1:{0:d}'.format(port_num)
        cls.srv.set_endpoint(cls.srv_url)
        objects = cls.srv.get_objects_node()
        obj = objects.add_object(4, "directory")
        var = obj.add_variable(4, "variable", 1.999)
        var2 = obj.add_variable(4, "variable2", 1.777)
        var2.set_writable()
        cls.srv.start()

    def test_uals(self):
        s = subprocess.check_output(["python", "../tools/uals", "--url", self.srv_url])
        self.assertIn(b"i=85", s)
        self.assertNotIn(b"i=89", s)
        self.assertNotIn(b"1.999", s)
        s = subprocess.check_output(["python", "../tools/uals", "--url", self.srv_url, "-d", "3"])
        self.assertIn(b"1.999", s)

    def test_uaread(self):
        s = subprocess.check_output(["python", "../tools/uaread", "--url", self.srv_url, "--path", "0:Objects,4:directory,4:variable"])
        self.assertIn(b"1.999", s)

    def test_uawrite(self):
        s = subprocess.check_output(["python", "../tools/uawrite", "--url", self.srv_url, "--path", "0:Objects,4:directory,4:variable2", "1.789"])
        s = subprocess.check_output(["python", "../tools/uaread", "--url", self.srv_url, "--path", "0:Objects,4:directory,4:variable2"])
        self.assertIn(b"1.789", s)
        self.assertNotIn(b"1.999", s)

    def test_uadiscover(self):
        s = subprocess.check_output(["python", "../tools/uadiscover", "--url", self.srv_url])
        self.assertIn(b"opc.tcp://127.0.0.1", s)
        self.assertIn(b"FreeOpcUa", s)
        self.assertIn(b"urn:freeopcua:python:server", s)

    @classmethod
    def tearDownClass(cls):
        cls.srv.stop()





import unittest

from opcua import Client
from opcua import Server
from opcua import ua

try:
    from opcua.crypto import uacrypto
    from opcua.crypto import security_policies
except ImportError:
    print("WARNING: CRYPTO NOT AVAILABLE, CRYPTO TESTS DISABLED!!")
    disable_crypto_tests = True
else:
    disable_crypto_tests = False


port_num1 = 48515
port_num2 = 48512
port_num3 = 48510
port_num4 = 48533

@unittest.skipIf(disable_crypto_tests, "crypto not available")
class TestCryptoConnect(unittest.TestCase):

    '''
    Test connectino with a server supporting crypto 

    '''
    @classmethod
    def setUpClass(cls):
        # start our own server
        cls.srv_crypto = Server()
        cls.uri_crypto = 'opc.tcp://127.0.0.1:{0:d}'.format(port_num1)
        cls.srv_crypto.set_endpoint(cls.uri_crypto)
        # load server certificate and private key. This enables endpoints
        # with signing and encryption.
        cls.srv_crypto.load_certificate("../examples/certificate-example.der")
        cls.srv_crypto.load_private_key("../examples/private-key-example.pem")
        cls.srv_crypto.start()

        # start a server without crypto
        cls.srv_no_crypto = Server()
        cls.uri_no_crypto = 'opc.tcp://127.0.0.1:{0:d}'.format(port_num2)
        cls.srv_no_crypto.set_endpoint(cls.uri_no_crypto)
        cls.srv_no_crypto.start()

        # start server with long key
        cls.srv_crypto2 = Server()
        cls.uri_crypto2 = 'opc.tcp://127.0.0.1:{0:d}'.format(port_num3)
        cls.srv_crypto2.set_endpoint(cls.uri_crypto2)
        cls.srv_crypto2.load_certificate("../examples/certificate-3072-example.der")
        cls.srv_crypto2.load_private_key("../examples/private-key-3072-example.pem")
        cls.srv_crypto2.start()

        cls.srv_crypto_no_anoymous = Server()
        cls.uri_crypto_no_anoymous = 'opc.tcp://127.0.0.1:{0:d}'.format(port_num4)
        cls.srv_crypto_no_anoymous.set_endpoint(cls.uri_crypto_no_anoymous)
        cls.srv_crypto_no_anoymous.load_certificate("examples/certificate-3072-example.der")
        cls.srv_crypto_no_anoymous.load_private_key("examples/private-key-3072-example.pem")
        cls.srv_crypto_no_anoymous.set_security_IDs(["Username", "Basic256Sha256"])
        cls.srv_crypto_no_anoymous.start()

    @classmethod
    def tearDownClass(cls):
        # stop the server 
        cls.srv_no_crypto.stop()
        cls.srv_crypto.stop()
        cls.srv_crypto2.stop()

    def test_nocrypto(self):
        clt = Client(self.uri_no_crypto)
        clt.connect()
        try:
            clt.get_objects_node().get_children()
        finally:
            clt.disconnect()

    def test_nocrypto_fail(self):
        clt = Client(self.uri_no_crypto)
        with self.assertRaises(ua.UaError):
            clt.set_security_string("Basic256Sha256,Sign,../examples/certificate-example.der,../examples/private-key-example.pem")

    def test_basic256sha256(self):
        clt = Client(self.uri_crypto)
        try:
            clt.set_security_string("Basic256Sha256,Sign,../examples/certificate-example.der,../examples/private-key-example.pem")
            clt.connect()
            self.assertTrue(clt.get_objects_node().get_children())
        finally:
            clt.disconnect()

    def test_basic256sha256_longkey(self):
        clt = Client(self.uri_crypto2)
        try:
            clt.set_security_string("Basic256Sha256,Sign,../examples/certificate-example.der,../examples/private-key-example.pem")
            clt.connect()
            self.assertTrue(clt.get_objects_node().get_children())
        finally:
            clt.disconnect()

    def test_basic256sha256_encrypt(self):
        clt = Client(self.uri_crypto)
        try:
            clt.set_security_string("Basic256Sha256,SignAndEncrypt,../examples/certificate-example.der,../examples/private-key-example.pem")
            clt.connect()
            self.assertTrue(clt.get_objects_node().get_children())
        finally:
            clt.disconnect()

    def test_basic256sha256_encrypt_longkey(self):
        clt = Client(self.uri_crypto2)
        try:
            clt.set_security_string("Basic256Sha256,SignAndEncrypt,../examples/certificate-example.der,../examples/private-key-example.pem")
            clt.connect()
            self.assertTrue(clt.get_objects_node().get_children())
        finally:
            clt.disconnect()

    def test_basic256sha56_encrypt_success(self):
        clt = Client(self.uri_crypto)
        try:
            clt.set_security(security_policies.SecurityPolicyBasic256Sha256,
                             '../examples/certificate-example.der',
                             '../examples/private-key-example.pem',
                             None,
                             ua.MessageSecurityMode.SignAndEncrypt
                             )
            clt.connect()
            self.assertTrue(clt.get_objects_node().get_children())
        finally:
            clt.disconnect()

    def test_basic256sha56_encrypt_fail(self):
        # FIXME: how to make it fail???
        clt = Client(self.uri_crypto)
        with self.assertRaises(ua.UaError):
            clt.set_security(security_policies.SecurityPolicyBasic256Sha256,
                             '../examples/certificate-example.der',
                             '../examples/private-key-example.pem',
                             None,
                             ua.MessageSecurityMode.None_
                             )

    def test_anonymous_rejection(self):
        # FIXME: how to make it fail???
        clt = Client(self.uri_crypto_no_anoymous)
        with self.assertRaises(ua.UaError) as exc_info:
            clt.connect()
            clt.disconnect()
        assert ua.StatusCodes.BadIdentityTokenRejected == exc_info.type.code



import unittest
import os.path
import xml.etree.ElementTree as ET

from opcua import ua
from opcua.server.address_space import AddressSpace
from opcua.server.address_space import NodeManagementService
from opcua.server.standard_address_space import standard_address_space

def find_elem(parent, name, ns = None):
    if ns is None:
        try:
            return parent.find(parent.tag[0:parent.tag.index('}')+1]+name)
        except ValueError:
            return parent.find(name)
    return parent.find(ns+name)

def remove_elem(parent, name):
    e = find_elem(parent, name)
    if e is not None:
        parent.remove(e)

def try_apply(item, aliases):
    attrib = item.attrib
    for name in ('ReferenceType', 'DataType'):
        try:
            value  = attrib[name]
            attrib[name] = aliases[value]
        except KeyError:
            pass

def read_nodes(path):
    tree = ET.parse(path)
    root = tree.getroot()

    aliases_elem = find_elem(root, 'Aliases')
    aliases = dict( (a.attrib['Alias'], a.text) for a in aliases_elem)

    any(try_apply(i, aliases) for i in root.iter())
    root.remove(aliases_elem)

    remove_elem(root, "Models")
    remove_elem(root, "NamespaceUris")

    return dict((e.attrib['NodeId'],e) for e in root)

def get_refs(e):
    return set((r.attrib['ReferenceType'], r.text, r.attrib.get('IsForward', 'true') == 'true') for r in find_elem(e, 'References'))

class StandardAddressSpaceTests(unittest.TestCase):

    def setUp(self):
        self.aspace = AddressSpace()
        self.node_mgt_service = NodeManagementService(self.aspace)
        standard_address_space.fill_address_space(self.node_mgt_service)

    # def test_std_address_space_references(self):
    #     std_nodes = read_nodes(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'schemas', 'Opc.Ua.NodeSet2.xml')))
    #     for k in self.aspace.keys():
    #         refs = set((r.ReferenceTypeId.to_string(), r.NodeId.to_string(), r.IsForward) for r in self.aspace[k].references)
    #         # FIXME: should renable this but it fails and I do not undertsand it
    #         #xml_refs = set((r.attrib['ReferenceType'], r.text, r.attrib.get('IsForward', 'true') == 'true') for r in find_elem(std_nodes[k.to_string()], 'References'))
    #         #self.assertTrue(len(xml_refs-refs)==0)
